from math import cos, radians, sin, tan
import importlib
import warnings

from OpenGL import GL
import numpy as np

from .. import Vector
from .. import functions as fn
from .. import getConfigOption
from ..Qt import QtCore, QtGui, QtWidgets, QT_LIB

if QT_LIB in ["PyQt5", "PySide2"]:
    QtOpenGL = QtGui
else:
    QtOpenGL = importlib.import_module(f"{QT_LIB}.QtOpenGL")

class GLViewMixin:
    def __init__(self, *args, rotationMethod='euler', **kwargs):
        """
        Mixin class providing functionality for GLViewWidget

        ================ ==============================================================
        **Arguments:**
        rotationMethod   (str): Mechanism to drive the rotation method, options are
                         'euler' and 'quaternion'. Defaults to 'euler'.
        ================ ==============================================================
        """
        super().__init__(*args, **kwargs)

        if rotationMethod not in ["euler", "quaternion"]:
            raise ValueError("Rotation method should be either 'euler' or 'quaternion'")
        
        self.opts = {
            'rotationMethod': rotationMethod
        }
        self.reset()
        self.items = []
        
        self.noRepeatKeys = [QtCore.Qt.Key.Key_Right, QtCore.Qt.Key.Key_Left, QtCore.Qt.Key.Key_Up, QtCore.Qt.Key.Key_Down, QtCore.Qt.Key.Key_PageUp, QtCore.Qt.Key.Key_PageDown]
        self.keysPressed = {}
        self.keyTimer = QtCore.QTimer()
        self.keyTimer.timeout.connect(self.evalKeyState)

        self._modelViewStack = []
        self._projectionStack = []
        self.default_vao = QtOpenGL.QOpenGLVertexArrayObject(self)

    def deviceWidth(self):
        dpr = self.devicePixelRatioF()
        return int(self.width() * dpr)

    def deviceHeight(self):
        dpr = self.devicePixelRatioF()
        return int(self.height() * dpr)

    def reset(self):
        """
        Initialize the widget state or reset the current state to the original state.
        """
        self.opts['center'] = Vector(0,0,0)  ## will always appear at the center of the widget
        self.opts['distance'] = 10.0         ## distance of camera from center
        self.opts['fov'] = 60                ## horizontal field of view in degrees

        if self.opts['rotationMethod'] == 'quaternion':
            self.opts['rotation'] = QtGui.QQuaternion(1,0,0,0)  ## camera rotation (quaternion:wxyz)
        else:
            self.opts['elevation'] = 30      ## camera's angle of elevation in degrees
            self.opts['azimuth'] = 45        ## camera's azimuthal angle in degrees
                                             ## (rotation around z-axis 0 points along x-axis)

        self.setBackgroundColor(getConfigOption('background'))

    def addItem(self, item):
        self.items.append(item)

        if self.isValid():
            item.initialize()
                
        item._setView(self)
        self.update()
        
    def removeItem(self, item):
        """
        Remove the item from the scene.
        """
        self.items.remove(item)
        item._setView(None)
        self.update()

    def clear(self):
        """
        Remove all items from the scene.
        """
        for item in self.items:
            item._setView(None)
        self.items = []
        self.update()        
        
    def initializeGL(self):
        """
        Initialize items that were not initialized during addItem().
        """
        ctx = self.context()
        fmt = ctx.format()
        if ctx.isOpenGLES():
            warnings.warn(
                f"pyqtgraph.opengl is primarily tested against OpenGL Desktop"
                f" but OpenGL {fmt.version()} ES detected",
                RuntimeWarning,
                stacklevel=2
            )
        elif fmt.version() < (2, 1):
            verString = GL.glGetString(GL.GL_VERSION)
            raise RuntimeError(
                "pyqtgraph.opengl: Requires >= OpenGL 2.1; Found %s" % verString
            )

        # Core profile requires a non-default VAO
        if fmt.profile() == QtGui.QSurfaceFormat.OpenGLContextProfile.CoreProfile:
            if not self.default_vao.isCreated():
                self.default_vao.create()
                self.default_vao.bind()

        for item in self.items:
            if not item.isInitialized():
                item.initialize()
        
    def setBackgroundColor(self, *args, **kwds):
        """
        Set the background color of the widget. Accepts the same arguments as
        :func:`~pyqtgraph.mkColor`.
        """
        self.opts['bgcolor'] = fn.mkColor(*args, **kwds).getRgbF()
        self.update()
        
    def getViewport(self):
        return (0, 0, self.width(), self.height())
        
    def setProjection(self, region, viewport):
        m = self.projectionMatrix(region, viewport)
        self._projectionStack.clear()
        self._projectionStack.append(m)

    def projectionMatrix(self, region, viewport):
        x0, y0, w, h = viewport
        dist = self.opts['distance']
        fov = self.opts['fov']
        nearClip = dist * 0.001
        farClip = dist * 1000.

        r = nearClip * tan(0.5 * radians(fov))
        t = r * h / w

        ## Note that X0 and width in these equations must be the values used in viewport
        left  = r * ((region[0]-x0) * (2.0/w) - 1)
        right = r * ((region[0]+region[2]-x0) * (2.0/w) - 1)
        bottom = t * ((region[1]-y0) * (2.0/h) - 1)
        top    = t * ((region[1]+region[3]-y0) * (2.0/h) - 1)

        tr = QtGui.QMatrix4x4()
        tr.frustum(left, right, bottom, top, nearClip, farClip)
        return tr
        
    def setModelview(self):
        m = self.viewMatrix()
        self._modelViewStack.clear()
        self._modelViewStack.append(m)
        
    def viewMatrix(self):
        tr = QtGui.QMatrix4x4()
        tr.translate( 0.0, 0.0, -self.opts['distance'])
        if self.opts['rotationMethod'] == 'quaternion':
            tr.rotate(self.opts['rotation'])
        else:
            # default rotation method
            tr.rotate(self.opts['elevation']-90, 1, 0, 0)
            tr.rotate(self.opts['azimuth']+90, 0, 0, -1)  
        center = self.opts['center']
        tr.translate(-center.x(), -center.y(), -center.z())
        return tr

    def currentModelView(self):
        return self._modelViewStack[-1]

    def currentProjection(self):
        return self._projectionStack[-1]

    def itemsAt(self, region=None):
        """
        Return a list of the items displayed in the region (x, y, w, h)
        relative to the widget.        
        """
        region = (region[0], self.height()-(region[1]+region[3]), region[2], region[3])
        viewport = self.getViewport()
        
        #buf = np.zeros(100000, dtype=np.uint)
        buf = GL.glSelectBuffer(100000)
        try:
            GL.glRenderMode(GL.GL_SELECT)
            GL.glInitNames()
            GL.glPushName(0)
            self._itemNames = {}
            self.paint(region=region, viewport=viewport, useItemNames=True)
            
        finally:
            hits = GL.glRenderMode(GL.GL_RENDER)
            
        items = [(h.near, h.names[0]) for h in hits]
        items.sort(key=lambda i: i[0])
        return [self._itemNames[i[1]] for i in items]
    
    def paintGL(self):
        # when called by Qt, glViewport has already been called
        # with device pixel ratio taken of
        region = self.getViewport()
        self.paint(region=region, viewport=region)

    def paint(self, *, region, viewport, useItemNames=False):
        """
        It is caller's responsibility to call glViewport prior to calling this method.
        region specifies the sub-region of viewport that should be rendered.
        """
        self.setProjection(region, viewport)
        self.setModelview()
        bgcolor = self.opts['bgcolor']
        GL.glClearColor(*bgcolor)
        GL.glClear( GL.GL_DEPTH_BUFFER_BIT | GL.GL_COLOR_BUFFER_BIT )
        self.drawItemTree(useItemNames=useItemNames)
        
    def drawItemTree(self, item=None, useItemNames=False):
        if item is None:
            items = [x for x in self.items if x.parentItem() is None]
        else:
            items = item.childItems()
            items.append(item)
        items.sort(key=lambda a: a.depthValue())
        for i in items:
            if not i.visible():
                continue
            if i is item:
                try:
                    if useItemNames:
                        GL.glLoadName(i._id)
                        self._itemNames[i._id] = i

                    # The GLGraphicsItem(s) making use of QPainter end
                    # up indirectly unbinding the default VAO, so we
                    # rebind it before each GLGraphicsItem.
                    if self.default_vao.isCreated():
                        self.default_vao.bind()

                    i.paint()
                except:
                    from .. import debug
                    debug.printExc()
                    print("Error while drawing item %s." % str(item))
            else:
                self._modelViewStack.append(self.currentModelView() * i.transform())
                try:
                    self.drawItemTree(i, useItemNames=useItemNames)
                finally:
                    self._modelViewStack.pop()

    def setCameraPosition(self, pos=None, distance=None, elevation=None, azimuth=None, rotation=None):
        if rotation is not None:
            # Alternatively, we could define that rotation overrides elevation and azimuth
            if elevation is not None:
                raise ValueError("cannot set both rotation and elevation")
            if azimuth is not None:
                raise ValueError("cannot set both rotation and azimuth")

        if pos is not None:
            self.opts['center'] = pos
        if distance is not None:
            self.opts['distance'] = distance

        if self.opts['rotationMethod'] == "quaternion":
            # note that "quaternion" mode modifies only opts['rotation']
            if elevation is not None or azimuth is not None:
                eu = self.opts['rotation'].toEulerAngles()
                if azimuth is not None:
                    eu.setZ(-azimuth-90)
                if elevation is not None:
                    eu.setX(elevation-90)
                self.opts['rotation'] = QtGui.QQuaternion.fromEulerAngles(eu)
            if rotation is not None:
                self.opts['rotation'] = rotation
        else:
            # note that "euler" mode modifies only opts['elevation'] and opts['azimuth']
            if elevation is not None:
                self.opts['elevation'] = elevation
            if azimuth is not None:
                self.opts['azimuth'] = azimuth
            if rotation is not None:
                eu = rotation.toEulerAngles()
                self.opts['elevation'] = eu.x() + 90
                self.opts['azimuth'] = -eu.z() - 90

        self.update()
        
    def cameraPosition(self):
        """Return current position of camera based on center, dist, elevation, and azimuth"""
        center = self.opts['center']
        dist = self.opts['distance']
        if self.opts['rotationMethod'] == "quaternion":
            pos = Vector(center - self.opts['rotation'].rotatedVector(Vector(0,0,dist) ))
        else:
            # using 'euler' rotation method
            elev = radians(self.opts['elevation'])
            azim = radians(self.opts['azimuth'])
            pos = Vector(
                center.x() + dist * cos(elev) * cos(azim),
                center.y() + dist * cos(elev) * sin(azim),
                center.z() + dist * sin(elev)
            )
        return pos

    def setCameraParams(self, **kwds):
        valid_keys = {'center', 'rotation', 'distance', 'fov', 'elevation', 'azimuth'}
        if not valid_keys.issuperset(kwds):
            raise ValueError(f'valid keywords are {valid_keys}')

        self.setCameraPosition(pos=kwds.get('center'), distance=kwds.get('distance'),
                               elevation=kwds.get('elevation'), azimuth=kwds.get('azimuth'),
                               rotation=kwds.get('rotation'))
        if 'fov' in kwds:
            self.opts['fov'] = kwds['fov']

    def cameraParams(self):
        valid_keys = ['center', 'distance', 'fov']

        if self.opts['rotationMethod'] == 'quaternion':
            valid_keys.append('rotation')
        else:
            valid_keys.extend(['elevation', 'azimuth'])

        return { k : self.opts[k] for k in valid_keys }

    def orbit(self, azim, elev):
        """Orbits the camera around the center position. *azim* and *elev* are given in degrees."""
        if self.opts['rotationMethod'] == 'quaternion':
            q = QtGui.QQuaternion.fromEulerAngles(
                    elev, -azim, 0
                    ) # rx-ry-rz
            q *= self.opts['rotation']
            self.opts['rotation'] = q
        else: # default euler rotation method
            self.opts['azimuth'] += azim
            self.opts['elevation'] = fn.clip_scalar(self.opts['elevation'] + elev, -90., 90.)
        self.update()
        
    def pan(self, dx, dy, dz, relative='global'):
        """
        Moves the center (look-at) position while holding the camera in place. 
        
        ==============  =======================================================
        **Arguments:**
        *dx*            Distance to pan in x direction
        *dy*            Distance to pan in y direction
        *dz*            Distance to pan in z direction
        *relative*      String that determines the direction of dx,dy,dz. 
                        If "global", then the global coordinate system is used.
                        If "view", then the z axis is aligned with the view
                        direction, and x and y axes are in the plane of the
                        view: +x points right, +y points up. 
                        If "view-upright", then x is in the global xy plane and
                        points to the right side of the view, y is in the
                        global xy plane and orthogonal to x, and z points in
                        the global z direction.
        ==============  =======================================================
        
        Distances are scaled roughly such that a value of 1.0 moves
        by one pixel on screen.
        """
        if relative == 'global':
            self.opts['center'] += QtGui.QVector3D(dx, dy, dz)
        elif relative == 'view-upright':
            cPos = self.cameraPosition()
            cVec = self.opts['center'] - cPos
            dist = cVec.length()  ## distance from camera to center
            xDist = dist * 2. * tan(0.5 * radians(self.opts['fov']))  ## approx. width of view at distance of center point
            xScale = xDist / self.width()
            zVec = QtGui.QVector3D(0,0,1)
            xVec = QtGui.QVector3D.crossProduct(zVec, cVec).normalized()
            yVec = QtGui.QVector3D.crossProduct(xVec, zVec).normalized()
            self.opts['center'] = self.opts['center'] + xVec * xScale * dx + yVec * xScale * dy + zVec * xScale * dz
        elif relative == 'view':
            # pan in plane of camera

            if self.opts['rotationMethod'] == 'quaternion':
                # obtain basis vectors
                qc = self.opts['rotation'].conjugated()
                xv = qc.rotatedVector( Vector(1,0,0) )
                yv = qc.rotatedVector( Vector(0,1,0) )
                zv = qc.rotatedVector( Vector(0,0,1) )

                scale_factor = self.pixelSize( self.opts['center'] )

                # apply translation
                self.opts['center'] += scale_factor * (xv*-dx + yv*dy + zv*dz)
            else: # use default euler rotation method
                elev = radians(self.opts['elevation'])
                azim = radians(self.opts['azimuth'])
                fov = radians(self.opts['fov'])
                dist = (self.opts['center'] - self.cameraPosition()).length()
                fov_factor = tan(fov / 2) * 2
                scale_factor = dist * fov_factor / self.width()
                z = scale_factor * cos(elev) * dy
                x = scale_factor * (sin(azim) * dx - sin(elev) * cos(azim) * dy)
                y = scale_factor * (cos(azim) * dx + sin(elev) * sin(azim) * dy)
                self.opts['center'] += QtGui.QVector3D(x, -y, z)
        else:
            raise ValueError("relative argument must be global, view, or view-upright")
        
        self.update()
        
    def pixelSize(self, pos):
        """
        Return the approximate size of a screen pixel at the location pos
        Pos may be a Vector or an (N,3) array of locations
        """
        cam = self.cameraPosition()
        if isinstance(pos, np.ndarray):
            cam = np.array(cam).reshape((1,)*(pos.ndim-1)+(3,))
            dist = ((pos-cam)**2).sum(axis=-1)**0.5
        else:
            dist = (pos-cam).length()
        xDist = dist * 2. * tan(0.5 * radians(self.opts['fov']))
        return xDist / self.width()

    def mousePressEvent(self, ev):
        lpos = ev.position() if hasattr(ev, 'position') else ev.localPos()
        self.mousePos = lpos

    def mouseMoveEvent(self, ev):
        lpos = ev.position() if hasattr(ev, 'position') else ev.localPos()
        if not hasattr(self, 'mousePos'):
            self.mousePos = lpos
        diff = lpos - self.mousePos
        self.mousePos = lpos
        
        if ev.buttons() == QtCore.Qt.MouseButton.LeftButton:
            if (ev.modifiers() & QtCore.Qt.KeyboardModifier.ControlModifier):
                self.pan(diff.x(), diff.y(), 0, relative='view')
            else:
                self.orbit(-diff.x(), diff.y())
        elif ev.buttons() == QtCore.Qt.MouseButton.MiddleButton:
            if (ev.modifiers() & QtCore.Qt.KeyboardModifier.ControlModifier):
                self.pan(diff.x(), 0, diff.y(), relative='view-upright')
            else:
                self.pan(diff.x(), diff.y(), 0, relative='view-upright')
        
    def mouseReleaseEvent(self, ev):
        pass
        
    def wheelEvent(self, ev):
        delta = ev.angleDelta().x()
        if delta == 0:
            delta = ev.angleDelta().y()
        if (ev.modifiers() & QtCore.Qt.KeyboardModifier.ControlModifier):
            self.opts['fov'] *= 0.999**delta
        else:
            self.opts['distance'] *= 0.999**delta
        self.update()

    def keyPressEvent(self, ev):
        if ev.key() in self.noRepeatKeys:
            ev.accept()
            if ev.isAutoRepeat():
                return
            self.keysPressed[ev.key()] = 1
            self.evalKeyState()
      
    def keyReleaseEvent(self, ev):
        if ev.key() in self.noRepeatKeys:
            ev.accept()
            if ev.isAutoRepeat():
                return
            try:
                del self.keysPressed[ev.key()]
            except KeyError:
                self.keysPressed = {}
            self.evalKeyState()
        
    def evalKeyState(self):
        speed = 2.0
        if len(self.keysPressed) > 0:
            for key in self.keysPressed:
                if key == QtCore.Qt.Key.Key_Right:
                    self.orbit(azim=-speed, elev=0)
                elif key == QtCore.Qt.Key.Key_Left:
                    self.orbit(azim=speed, elev=0)
                elif key == QtCore.Qt.Key.Key_Up:
                    self.orbit(azim=0, elev=-speed)
                elif key == QtCore.Qt.Key.Key_Down:
                    self.orbit(azim=0, elev=speed)
                elif key == QtCore.Qt.Key.Key_PageUp:
                    pass
                elif key == QtCore.Qt.Key.Key_PageDown:
                    pass
                self.keyTimer.start(16)
        else:
            self.keyTimer.stop()

    def readQImage(self):
        """
        Read the current buffer pixels out as a QImage.
        """
        return self.grabFramebuffer()
        
    def renderToArray(self, size, format=GL.GL_BGRA, type=GL.GL_UNSIGNED_BYTE, textureSize=1024, padding=256):
        w,h = map(int, size)
        
        self.makeCurrent()

        texwidth = textureSize

        fbo = QtOpenGL.QOpenGLFramebufferObject(texwidth, texwidth,
                    QtOpenGL.QOpenGLFramebufferObject.Attachment.CombinedDepthStencil,
                    GL.GL_TEXTURE_2D)

        output = np.empty((h, w, 4), dtype=np.ubyte)
        data = np.empty((texwidth, texwidth, 4), dtype=np.ubyte)

        try:
            p2 = 2 * padding
            for x in range(-padding, w-padding, texwidth-p2):
                for y in range(-padding, h-padding, texwidth-p2):
                    x2 = min(x+texwidth, w+padding)
                    y2 = min(y+texwidth, h+padding)
                    w2 = x2-x
                    h2 = y2-y
                    
                    fbo.bind()
                    GL.glViewport(0, 0, w2, h2)
                    self.paint(region=(x, h-y-h2, w2, h2), viewport=(0, 0, w, h))  # only render sub-region
                    
                    fbo.bind()
                    GL.glReadPixels(0, 0, texwidth, texwidth, format, type, data)
                    data_yflip = data[::-1, ...]
                    output[y+padding:y2-padding, x+padding:x2-padding] = data_yflip[-(h2-padding):-padding, padding:w2-padding]
                    
        finally:
            fbo.release()

        return output


class GLViewWidget(GLViewMixin, QtWidgets.QOpenGLWidget):
    def __init__(self, *args, devicePixelRatio=None, **kwargs):
        """
        Basic widget for displaying 3D data
          - Rotation/scale controls
          - Axis/grid display
          - Export options

        ================ ==============================================================
        **Arguments:**
        parent           (QObject, optional): Parent QObject. Defaults to None.
        devicePixelRatio No longer in use. High-DPI displays should automatically
                         detect the correct resolution.
        rotationMethod   (str): Mechanism to drive the rotation method, options are
                         'euler' and 'quaternion'. Defaults to 'euler'.
        ================ ==============================================================
        """
        super().__init__(*args, **kwargs)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.ClickFocus)



"""
RawImageWidget.py
Copyright 2010-2016 Luke Campagnola
Distributed under MIT/X11 license. See license.txt for more information.
"""

import importlib
import warnings
import numpy as np

from .. import functions as fn
from .. import functions_qimage
from .. import getConfigOption, getCupy
from ..Qt import QtCore, QtGui, QtWidgets, QT_LIB
from ..Qt import OpenGLConstants as GLC
from ..Qt import OpenGLHelpers

if QT_LIB in ["PyQt5", "PySide2"]:
    QtOpenGL = QtGui
    QtOpenGLWidgets = QtWidgets
else:
    QtOpenGL = importlib.import_module(f"{QT_LIB}.QtOpenGL")
    QtOpenGLWidgets = importlib.import_module(f"{QT_LIB}.QtOpenGLWidgets")

# importing cuda python is fast
try:
    from cuda.bindings import runtime as cudart
except ImportError:
    cudart = None

__all__ = ['RawImageWidget', 'RawImageGLWidget']

class RawImageWidget(QtWidgets.QWidget):
    """
    Widget optimized for very fast video display.
    Generally using an ImageItem inside GraphicsView is fast enough.
    On some systems this may provide faster video. See the VideoSpeedTest example for benchmarking.
    """

    def __init__(self, parent=None, scaled=False):
        """
        Setting scaled=True will cause the entire image to be displayed within the boundaries of the widget.
        This also greatly reduces the speed at which it will draw frames.
        """
        QtWidgets.QWidget.__init__(self, parent)
        self.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding))
        self.scaled = scaled
        self.opts = None
        self.image = None
        self._cp = getCupy()

    def setImage(self, img, *args, **kargs):
        """
        img must be ndarray of shape (x,y), (x,y,3), or (x,y,4).
        Extra arguments are sent to functions.makeARGB
        """
        if getConfigOption('imageAxisOrder') == 'col-major':
            img = img.swapaxes(0, 1)
        self.opts = (img, args, kargs)
        self.image = None
        self.update()

    def paintEvent(self, ev):
        if self.opts is None:
            return
        if self.image is None:
            img = self.opts[0]
            xp = self._cp.get_array_module(img) if self._cp else np

            qimage = None
            if (
                not self.opts[1]    # no positional arguments
                and {"levels", "lut"}.issuperset(self.opts[2])  # no kwargs besides levels and lut
            ):
                transparentLocations = None
                if img.dtype.kind == "f" and xp.isnan(img.min()):
                    nanmask = xp.isnan(img)
                    if nanmask.ndim == 3:
                        nanmask = nanmask.any(axis=2)
                    transparentLocations = nanmask.nonzero()

                qimage = functions_qimage.try_make_qimage(
                    img,
                    levels=self.opts[2].get("levels"),
                    lut=self.opts[2].get("lut"),
                    transparentLocations=transparentLocations
                )

            if qimage is None:
                argb, alpha = fn.makeARGB(self.opts[0], *self.opts[1], **self.opts[2])
                if self._cp and self._cp.get_array_module(argb) == self._cp:
                    argb = argb.get()  # transfer GPU data back to the CPU
                qimage = fn.ndarray_to_qimage(argb, QtGui.QImage.Format.Format_ARGB32)

            self.image = qimage
            self.opts = ()
        # if self.pixmap is None:
            # self.pixmap = QtGui.QPixmap.fromImage(self.image)
        p = QtGui.QPainter(self)
        if self.scaled:
            rect = self.rect()
            ar = rect.width() / float(rect.height())
            imar = self.image.width() / float(self.image.height())
            if ar > imar:
                rect.setWidth(int(rect.width() * imar / ar))
            else:
                rect.setHeight(int(rect.height() * ar / imar))

            p.drawImage(rect, self.image)
        else:
            p.drawImage(QtCore.QPointF(), self.image)
        # p.drawPixmap(self.rect(), self.pixmap)
        p.end()

def checkCudaErrors(result):
    if err_num := result[0]:
        error_name = cudart.cudaGetErrorName(err_num)[1].decode()
        raise RuntimeError(f"{error_name}({err_num})")
    if len(result) == 1:
        return None
    elif len(result) == 2:
        return result[1]
    else:
        return result[1:]

class RawImageGLWidget(QtOpenGLWidgets.QOpenGLWidget):
    """
    Similar to RawImageWidget, but uses a GL widget to do all drawing.
    Performance varies between platforms; see examples/VideoSpeedTest for benchmarking.

    Checks if setConfigOptions(imageAxisOrder='row-major') was set.
    """

    def __init__(self, parent=None, smooth=False):
        super().__init__(parent)
        self.image = None
        self.uploaded = False
        self.smooth = smooth
        self.opts = None
        self.gfx_resource = None

        self.m_texture = QtOpenGL.QOpenGLTexture(QtOpenGL.QOpenGLTexture.Target.Target2D)
        self.m_blitter = QtOpenGL.QOpenGLTextureBlitter()

        self.try_cuda = cudart is not None

    def setImage(self, img, *args, **kargs):
        """
        img must be ndarray of shape (x,y), (x,y,3), or (x,y,4).
        Extra arguments are sent to functions.makeARGB
        """
        if getConfigOption('imageAxisOrder') == 'col-major':
            img = img.swapaxes(0, 1)
        self.opts = (img, args, kargs)
        self.image = None
        self.uploaded = False
        self.update()

    def initializeGL(self):
        ctx = self.context()

        # in Python, slot will not get called during application termination
        ctx.aboutToBeDestroyed.connect(self.cleanup)

        self.glfn = OpenGLHelpers.getFunctions(ctx)

        self.m_blitter.create()

    def cleanup(self):
        # explicit call of cleanup() is needed during application termination
        self.makeCurrent()
        if self.gfx_resource is not None:
            cudart.cudaGraphicsUnregisterResource(self.gfx_resource)
            self.gfx_resource = None
        self.m_texture.destroy()
        self.m_blitter.destroy()
        self.uploaded = False
        self.doneCurrent()

    def uploadTexture(self):
        rgba = self.image
        h, w = rgba.shape[:2]

        if self.m_texture.isCreated() and (w != self.m_texture.width() or h != self.m_texture.height()):
            if self.gfx_resource is not None:
                cudart.cudaGraphicsUnregisterResource(self.gfx_resource)
                self.gfx_resource = None
            self.m_texture.destroy()

        if not self.m_texture.isCreated():
            self.m_texture.setFormat(QtOpenGL.QOpenGLTexture.TextureFormat.RGBA8_UNorm)
            self.m_texture.setSize(w, h)
            self.m_texture.allocateStorage()

        filt = QtOpenGL.QOpenGLTexture.Filter.Linear if self.smooth else QtOpenGL.QOpenGLTexture.Filter.Nearest
        self.m_texture.setMinMagFilters(filt, filt)
        self.m_texture.setWrapMode(QtOpenGL.QOpenGLTexture.WrapMode.ClampToBorder)

        # cupy, pytorch and numba support CAI
        data_on_gpu = hasattr(rgba, '__cuda_array_interface__')

        if data_on_gpu and self.try_cuda and self.gfx_resource is None:
            try:
                self.gfx_resource = checkCudaErrors(cudart.cudaGraphicsGLRegisterImage(
                    self.m_texture.textureId(), self.m_texture.target().value,
                    cudart.cudaGraphicsRegisterFlags.cudaGraphicsRegisterFlagsWriteDiscard
                ))
            except RuntimeError as e:
                # registration can fail if the OpenGL gpu is not the CUDA gpu,
                warnings.warn(f"cudaGraphicsGLRegisterImage failed: {e}")
                # registration failed, don't try to register anymore
                self.try_cuda = False

        if data_on_gpu and self.gfx_resource is None:
            # cannot do gpu direct transfer, so transfer back to cpu
            rgba = rgba.get()   # cupy syntax
            data_on_gpu = False

        if not data_on_gpu:
            self.m_texture.setData(
                QtOpenGL.QOpenGLTexture.PixelFormat.RGBA,
                QtOpenGL.QOpenGLTexture.PixelType.UInt8,
                rgba)
        else:
            cai = rgba.__cuda_array_interface__
            device_ptr = cai['data'][0]
            stream = cai.get('stream')

            checkCudaErrors(cudart.cudaGraphicsMapResources(1, self.gfx_resource, stream))
            try:
                cuda_array = checkCudaErrors(cudart.cudaGraphicsSubResourceGetMappedArray(
                    self.gfx_resource, 0, 0
                ))

                src_pitch = w * 4 * rgba.itemsize   # assume contiguous
                checkCudaErrors(cudart.cudaMemcpy2DToArrayAsync(
                    cuda_array, 0, 0,
                    device_ptr, src_pitch,
                    w * 4 * rgba.itemsize, h,
                    cudart.cudaMemcpyKind.cudaMemcpyDefault,
                    stream
                ))
            finally:
                cudart.cudaGraphicsUnmapResources(1, self.gfx_resource, stream)

    def paintGL(self):
        self.glfn.glClearColor(1, 1, 1, 1)
        self.glfn.glClear(GLC.GL_COLOR_BUFFER_BIT)
        self.glfn.glEnable(GLC.GL_BLEND)
        self.glfn.glBlendFuncSeparate(GLC.GL_SRC_ALPHA, GLC.GL_ONE_MINUS_SRC_ALPHA, 1, GLC.GL_ONE_MINUS_SRC_ALPHA)

        if self.image is None:
            if self.opts is None:
                return
            img, args, kwds = self.opts
            self.image, _ = fn.makeRGBA(img, *args, **kwds)

        if not self.uploaded:
            # mark as uploaded whether or not it succeeds so that we don't retry and refail
            self.uploaded = True
            self.uploadTexture()

        target = QtGui.QMatrix4x4()
        self.m_blitter.bind()
        self.m_blitter.blit(self.m_texture.textureId(), target, QtOpenGL.QOpenGLTextureBlitter.Origin.OriginTopLeft)
        self.m_blitter.release()



import enum
import importlib

import numpy as np

from .. import Qt, colormap
from .. import functions as fn
from .. import getConfigOption
from ..Qt import compat
from ..Qt import OpenGLConstants as GLC
from ..Qt import OpenGLHelpers
from ..Qt import QtCore, QtGui, QT_LIB
from .GraphicsObject import GraphicsObject

if QT_LIB in ["PyQt5", "PySide2"]:
    QtOpenGL = QtGui
else:
    QtOpenGL = importlib.import_module(f'{QT_LIB}.QtOpenGL')

__all__ = ['PColorMeshItem']


class DirtyFlag(enum.Flag):
    XY = enum.auto()
    Z = enum.auto()
    LUT = enum.auto()
    DIM = enum.auto()


class QuadInstances:
    def __init__(self):
        self.nrows = -1
        self.ncols = -1
        self.pointsarray = Qt.internals.PrimitiveArray(QtCore.QPointF, 2)
        self.resize(0, 0)

    def resize(self, nrows, ncols):
        if nrows == self.nrows and ncols == self.ncols:
            return

        self.nrows = nrows
        self.ncols = ncols

        # (nrows + 1) * (ncols + 1) vertices, (x, y)
        self.pointsarray.resize((nrows+1)*(ncols+1))
        points = self.pointsarray.instances()
        # points is a flattened list of a 2d array of
        # QPointF(s) of shape (nrows+1, ncols+1)

        # pre-create quads from those instances of QPointF(s).
        # store the quads as a flattened list of a 2d array
        # of polygons of shape (nrows, ncols)
        polys = np.ndarray(nrows*ncols, dtype=object)
        for r in range(nrows):
            for c in range(ncols):
                bl = points[(r+0)*(ncols+1)+(c+0)]
                tl = points[(r+0)*(ncols+1)+(c+1)]
                br = points[(r+1)*(ncols+1)+(c+0)]
                tr = points[(r+1)*(ncols+1)+(c+1)]
                poly = (bl, br, tr, tl)
                polys[r*ncols+c] = poly
        self.polys = polys

    def ndarray(self):
        return self.pointsarray.ndarray()

    def instances(self):
        return self.polys


class PColorMeshItem(GraphicsObject):
    """
    **Bases:** :class:`GraphicsObject <pyqtgraph.GraphicsObject>`
    """

    sigLevelsChanged = QtCore.Signal(object)  # emits tuple with levels (low,high) when color levels are changed.

    def __init__(self, *args, **kwargs):
        """
        Create a pseudocolor plot with convex polygons.

        Call signature:

        ``PColorMeshItem([x, y,] z, **kwargs)``

        x and y can be used to specify the corners of the quadrilaterals.
        z must be used to specified to color of the quadrilaterals.

        Parameters
        ----------
        x, y : np.ndarray, optional, default None
            2D array containing the coordinates of the polygons
        z : np.ndarray
            2D array containing the value which will be mapped into the polygons
            colors.
            If x and y is None, the polygons will be displaced on a grid
            otherwise x and y will be used as polygons vertices coordinates as::

                (x[i+1, j], y[i+1, j])           (x[i+1, j+1], y[i+1, j+1])
                                    +---------+
                                    | z[i, j] |
                                    +---------+
                    (x[i, j], y[i, j])           (x[i, j+1], y[i, j+1])

            "ASCII from: <https://matplotlib.org/3.2.1/api/_as_gen/matplotlib.pyplot.pcolormesh.html>".
        colorMap : pyqtgraph.ColorMap
            Colormap used to map the z value to colors.
            default ``pyqtgraph.colormap.get('viridis')``
        levels: tuple, optional, default None
            Sets the minimum and maximum values to be represented by the colormap (min, max).
            Values outside this range will be clipped to the colors representing min or max.
            ``None`` disables the limits, meaning that the colormap will autoscale
            the next time ``setData()`` is called with new data.
        enableAutoLevels: bool, optional, default True
            Causes the colormap levels to autoscale whenever ``setData()`` is called.
            It is possible to override this value on a per-change-basis by using the
            ``autoLevels`` keyword argument when calling ``setData()``.
            If ``enableAutoLevels==False`` and ``levels==None``, autoscaling will be
            performed once when the first z data is supplied.
        edgecolors : dict, optional
            The color of the edges of the polygons.
            Default None means no edges.
            Only cosmetic pens are supported.
            The dict may contains any arguments accepted by :func:`mkColor() <pyqtgraph.mkColor>`.
            Example: ``mkPen(color='w', width=2)``
        antialiasing : bool, default False
            Whether to draw edgelines with antialiasing.
            Note that if edgecolors is None, antialiasing is always False.
        """

        GraphicsObject.__init__(self)

        self.qpicture = None  ## rendered picture for display
        self.x = None
        self.y = None
        self.z = None
        self._dataBounds = None
        self.glstate = None

        self.edgecolors = kwargs.get('edgecolors', None)
        if self.edgecolors is not None:
            self.edgecolors = fn.mkPen(self.edgecolors)
            # force the pen to be cosmetic. see discussion in
            # https://github.com/pyqtgraph/pyqtgraph/pull/2586
            self.edgecolors.setCosmetic(True)
        self.antialiasing = kwargs.get('antialiasing', False)
        self.levels = kwargs.get('levels', None)
        self._defaultAutoLevels = kwargs.get('enableAutoLevels', True)

        if 'colorMap' in kwargs:
            cmap = kwargs.get('colorMap')
            if not isinstance(cmap, colormap.ColorMap):
                raise ValueError('colorMap argument must be a ColorMap instance')
            self.cmap = cmap
        else:
            self.cmap = colormap.get('viridis')

        self.lut_qcolor = self.cmap.getLookupTable(nPts=256, mode=self.cmap.QCOLOR)

        self.quads = QuadInstances()

        # If some data have been sent we directly display it
        if len(args)>0:
            self.setData(*args)


    def _prepareData(self, args) -> DirtyFlag:
        """
        Check the shape of the data.
        Return a set of 2d array x, y, z ready to be used to draw the picture.
        """

        dirtyFlags = DirtyFlag.XY | DirtyFlag.Z | DirtyFlag.DIM

        # User didn't specified data
        if len(args)==0:

            self.x = None
            self.y = None
            self.z = None

            self._dataBounds = None

        # User only specified z
        elif len(args)==1:
            # If x and y is None, the polygons will be displaced on a grid
            x = np.arange(0, args[0].shape[0]+1, 1)
            y = np.arange(0, args[0].shape[1]+1, 1)
            self.x, self.y = np.meshgrid(x, y, indexing='ij')
            self.z = args[0]

            self._dataBounds = ((x[0], x[-1]), (y[0], y[-1]))

        # User specified x, y, z
        elif len(args)==3:
            # specifying None explicitly means to retain the existing value
            if (x := args[0]) is None:
                x = self.x
            if (y := args[1]) is None:
                y = self.y
            if (z := args[2]) is None:
                z = self.z

            if args[0] is None and args[1] is None:
                dirtyFlags &= ~DirtyFlag.XY
            if args[2] is None:
                dirtyFlags &= ~DirtyFlag.Z

            if self.z is not None and z.shape == self.z.shape:
                dirtyFlags &= ~DirtyFlag.DIM

            # Shape checking
            xy_shape = (z.shape[0]+1, z.shape[1]+1)
            if x.shape != xy_shape:
                raise ValueError('The dimension of x should be one greater than the one of z')
            if y.shape != xy_shape:
                raise ValueError('The dimension of y should be one greater than the one of z')

            self.x = x
            self.y = y
            self.z = z

            xmn, xmx = np.min(self.x), np.max(self.x)
            ymn, ymx = np.min(self.y), np.max(self.y)
            self._dataBounds = ((xmn, xmx), (ymn, ymx))

        else:
            raise ValueError('Data must been sent as (z) or (x, y, z)')

        return dirtyFlags

    def setData(self, *args, **kwargs):
        """
        Set the data to be drawn.

        Parameters
        ----------
        x, y : np.ndarray, optional, default None
            2D array containing the coordinates of the polygons
        z : np.ndarray
            2D array containing the value which will be mapped into the polygons
            colors.
            If x and y is None, the polygons will be displaced on a grid
            otherwise x and y will be used as polygons vertices coordinates as::

                (x[i+1, j], y[i+1, j])           (x[i+1, j+1], y[i+1, j+1])
                                    +---------+
                                    | z[i, j] |
                                    +---------+
                    (x[i, j], y[i, j])           (x[i, j+1], y[i, j+1])

            "ASCII from: <https://matplotlib.org/3.2.1/api/_as_gen/
                         matplotlib.pyplot.pcolormesh.html>".
        autoLevels: bool, optional
            If set, overrides the value of ``enableAutoLevels``
        """
        old_bounds = self._dataBounds
        dirtyFlags = self._prepareData(args)
        boundsChanged = old_bounds != self._dataBounds

        self._rerender(
            autoLevels=kwargs.get('autoLevels', self._defaultAutoLevels)
        )

        if boundsChanged:
            self.prepareGeometryChange()
            self.informViewBoundsChanged()

        if self.glstate is not None:
            self.glstate.dataChange(dirtyFlags)

        self.update()

    def _rerender(self, *, autoLevels):
        self.qpicture = None
        if self.z is not None and np.any(np.isfinite(self.z)):
            if (self.levels is None) or autoLevels:
                # Autoscale colormap
                z_min = np.nanmin(self.z)
                z_max = np.nanmax(self.z)
                self.setLevels( (z_min, z_max), update=False)

    def _drawPicture(self) -> QtGui.QPicture:
        # on entry, the following members are all valid: x, y, z, levels
        # this function does not alter any state (besides using self.quads)

        picture = QtGui.QPicture()
        painter = QtGui.QPainter(picture)
        # We set the pen of all polygons once
        if self.edgecolors is None:
            painter.setPen(QtCore.Qt.PenStyle.NoPen)
        else:
            painter.setPen(self.edgecolors)
            if self.antialiasing:
                painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)

        z_invalid = np.isnan(self.z)
        skip_nans = np.any(z_invalid)
        if skip_nans:
            # note: flattens array
            valid_z = self.z[~z_invalid]
            if len(valid_z) == 0:
                # nothing to draw => return
                painter.end()
                return picture
        else:
            valid_z = self.z

        ## Prepare colormap
        # First we get the LookupTable
        lut = self.lut_qcolor
        # Second we associate each z value, that we normalize, to the lut
        scale = len(lut) - 1
        lo, hi = self.levels[0], self.levels[1]
        rng = hi - lo
        if rng == 0:
            rng = 1
        norm = fn.rescaleData(valid_z, scale / rng, lo, dtype=int, clip=(0, len(lut)-1))

        if Qt.QT_LIB.startswith('PyQt'):
            drawConvexPolygon = lambda x : painter.drawConvexPolygon(*x)
        else:
            drawConvexPolygon = painter.drawConvexPolygon

        self.quads.resize(self.z.shape[0], self.z.shape[1])
        memory = self.quads.ndarray()
        memory[..., 0] = self.x.ravel()
        memory[..., 1] = self.y.ravel()
        polys = self.quads.instances()

        if skip_nans:
            polys = polys[(~z_invalid).flat]

        # group indices of same coloridx together
        color_indices, counts = np.unique(norm, return_counts=True)
        # note: returns flattened array
        sorted_indices = np.argsort(norm, axis=None)

        offset = 0
        for coloridx, cnt in zip(color_indices, counts):
            indices = sorted_indices[offset:offset+cnt]
            offset += cnt
            painter.setBrush(lut[coloridx])
            for idx in indices:
                drawConvexPolygon(polys[idx])

        painter.end()
        return picture

    def setLevels(self, levels, update=True):
        """
        Sets color-scaling levels for the mesh.

        Parameters
        ----------
            levels: tuple
                ``(low, high)``
                sets the range for which values can be represented in the colormap.
            update: bool, optional
                Controls if mesh immediately updates to reflect the new color levels.
        """
        self.levels = levels
        self.sigLevelsChanged.emit(levels)
        if update:
            self._rerender(autoLevels=False)
            self.update()

    def getLevels(self):
        """
        Returns a tuple containing the current level settings. See :func:`~setLevels`.
        The format is ``(low, high)``.
        """
        return self.levels

    def setLookupTable(self, lut, update=True):
        self.cmap = None    # invalidate since no longer consistent with lut
        self.lut_qcolor = lut[:]
        if self.glstate is not None:
            self.glstate.dataChange(DirtyFlag.LUT)
        if update:
            self._rerender(autoLevels=False)
            self.update()

    def getColorMap(self):
        return self.cmap

    def setColorMap(self, cmap):
        self.setLookupTable(cmap.getLookupTable(nPts=256, mode=cmap.QCOLOR), update=True)
        self.cmap = cmap

    def enableAutoLevels(self):
        self._defaultAutoLevels = True

    def disableAutoLevels(self):
        self._defaultAutoLevels = False

    def paint(self, painter, opt, widget):
        if self.z is None:
            return

        if (
            isinstance(widget, OpenGLHelpers.GraphicsViewGLWidget)
            and self.cmap is not None   # don't support setting colormap by setLookupTable
        ):
            if self.glstate is None:
                self.glstate = OpenGLState(widget)
            painter.beginNativePainting()
            try:
                self.paintGL(widget)
            finally:
                painter.endNativePainting()

            if (
                self.edgecolors is not None
                and self.edgecolors.style() != QtCore.Qt.PenStyle.NoPen
            ):
                painter.setPen(self.edgecolors)
                if self.antialiasing:
                    painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)
                for idx in range(self.x.shape[0]):
                    painter.drawPolyline(fn.arrayToQPolygonF(self.x[idx, :], self.y[idx, :]))
                for idx in range(self.x.shape[1]):
                    painter.drawPolyline(fn.arrayToQPolygonF(self.x[:, idx], self.y[:, idx]))

            return

        if self.qpicture is None:
            self.qpicture = self._drawPicture()
        painter.drawPicture(0, 0, self.qpicture)

    def width(self):
        if self._dataBounds is None:
            return 0
        bounds = self._dataBounds[0]
        return bounds[1]-bounds[0]

    def height(self):
        if self._dataBounds is None:
            return 0
        bounds = self._dataBounds[1]
        return bounds[1]-bounds[0]

    def dataBounds(self, ax, frac=1.0, orthoRange=None):
        if self._dataBounds is None:
            return (None, None)
        return self._dataBounds[ax]

    def pixelPadding(self):
        # pen is known to be cosmetic
        pen = self.edgecolors
        no_pen = (pen is None) or (pen.style() == QtCore.Qt.PenStyle.NoPen)
        return 0 if no_pen else (pen.widthF() or 1) * 0.5

    def boundingRect(self):
        xmn, xmx = self.dataBounds(ax=0)
        if xmn is None or xmx is None:
            return QtCore.QRectF()
        ymn, ymx = self.dataBounds(ax=1)
        if ymn is None or ymx is None:
            return QtCore.QRectF()

        px = py = 0
        pxPad = self.pixelPadding()
        if pxPad > 0:
            # determine length of pixel in local x, y directions
            px, py = self.pixelVectors()
            px = 0 if px is None else px.length()
            py = 0 if py is None else py.length()
            # return bounds expanded by pixel size
            px *= pxPad
            py *= pxPad

        return QtCore.QRectF(xmn-px, ymn-py, (2*px)+xmx-xmn, (2*py)+ymx-ymn)

    def paintGL(self, widget):
        if (view := self.getViewBox()) is None:
            return

        X, Y, Z = self.x, self.y, self.z

        glstate = self.glstate
        glstate.setup(widget.context())
        glfn = widget.getFunctions()
        program = widget.retrieveProgram("PColorMeshItem")

        # OpenGL only sees the float32 version of our data, and this may cause
        # precision issues. To mitigate this, we shift the origin of our data
        # to the center of its bounds.
        # Note that xc, yc are double precision Python floats. Subtracting them
        # from the x, y ndarrays will automatically upcast the latter to double
        # precision.
        if glstate.render_cache is None:
            origin = None
            dirty_bits = DirtyFlag.XY | DirtyFlag.Z | DirtyFlag.LUT | DirtyFlag.DIM
        else:
            origin, dirty_bits = glstate.render_cache
        if origin is None or DirtyFlag.XY in dirty_bits:
            # the origin point is calculated once per data change.
            # once the data is uploaded, the origin point is fixed.
            center = self.boundingRect().center()
            origin = center.x(), center.y()

        proj = QtGui.QMatrix4x4()
        proj.ortho(widget.rect())
        tr = self.sceneTransform()
        xc, yc = origin
        tr.translate(xc, yc)
        mvp = proj * QtGui.QMatrix4x4(tr)

        if glstate.flat_shading:
            vtx_array_shape = X.shape
        else:
            vtx_array_shape = Z.shape + (4,)
        num_vtx_mesh = np.prod(vtx_array_shape)
        num_ind_mesh = np.prod(Z.shape) * 6

        # resize (and invalidate) gpu buffers if needed.
        # a reallocation can only occur together with a change in data.
        # i.e. reallocation ==> change in data (render_cache is None)
        if DirtyFlag.DIM in dirty_bits:
            glstate.m_vbo_pos.bind()
            glstate.m_vbo_pos.allocate(num_vtx_mesh * 2 * 4)
            glstate.m_vbo_pos.release()
            glstate.m_vbo_lum.bind()
            glstate.m_vbo_lum.allocate(num_vtx_mesh * 1 * 4)
            glstate.m_vbo_lum.release()

            if glstate.flat_shading:
                # let the bottom-left of each quad be its "anchor".
                # then each quad is made up of 2 triangles
                #   (TR, TL, BL); (BR, TR, BL)
                # that have indices
                #   (stride + 1, stride + 0, 0); (1, stride + 1, 0)
                # where "0" is the relative index of BL
                # and "stride" advances to the next row
                # note that both triangles are created such that their 3rd vertex is at "BL"
                stride = Z.shape[1] + 1
                dim0 = np.arange(0, Z.shape[0]*stride, stride, dtype=np.uint32)[:, np.newaxis, np.newaxis]
                dim1 = np.arange(Z.shape[1], dtype=np.uint32)[np.newaxis, :, np.newaxis]
                dim2 = np.array([stride + 1, stride + 0, 0, 1, stride + 1, 0], dtype=np.uint32)[np.newaxis, np.newaxis, :]
                buf_ind = dim0 + dim1 + dim2
            else:
                # for each quad, we store 4 vertices contiguously (BL, BR, TL, TR)
                # then each quad is made up of 2 triangles
                #   (TR, TL, BL); (BR, TR, BL)
                # that have indices
                #   (3, 2, 0); (1, 3, 0)
                strides = np.cumprod(vtx_array_shape[::-1])[::-1]
                dim0 = np.arange(0, strides[0], strides[1], dtype=np.uint32)[:, np.newaxis, np.newaxis]
                dim1 = np.arange(0, strides[1], strides[2], dtype=np.uint32)[np.newaxis, :, np.newaxis]
                dim2 = np.array([3, 2, 0, 1, 3, 0], dtype=np.uint32)[np.newaxis, np.newaxis, :]
                buf_ind = dim0 + dim1 + dim2

            glstate.m_vbo_ind.bind()
            glstate.m_vbo_ind.allocate(buf_ind, buf_ind.nbytes)
            glstate.m_vbo_ind.release()

            dirty_bits &= ~DirtyFlag.DIM

        if DirtyFlag.LUT in dirty_bits:
            lut = self.cmap.getLookupTable(nPts=256, alpha=True)
            glstate.setTextureLut(lut)
            dirty_bits &= ~DirtyFlag.LUT

        if DirtyFlag.XY in dirty_bits:
            pos = np.empty(vtx_array_shape + (2,), dtype=np.float32)

            if glstate.flat_shading:
                pos[..., 0] = X - xc
                pos[..., 1] = Y - yc
            else:
                XY = np.dstack((X - xc, Y - yc)).astype(np.float32)
                pos[..., 0, :] = XY[:-1, :-1, :] # BL
                pos[..., 1, :] = XY[1:, :-1, :]  # BR
                pos[..., 2, :] = XY[:-1, 1:, :]  # TL
                pos[..., 3, :] = XY[1:, 1:, :]   # TR

            glstate.m_vbo_pos.bind()
            glstate.m_vbo_pos.write(0, pos, pos.nbytes)
            glstate.m_vbo_pos.release()

            dirty_bits &= ~DirtyFlag.XY

        if DirtyFlag.Z in dirty_bits:
            lum = np.empty(vtx_array_shape, dtype=np.float32)

            if glstate.flat_shading:
                lum[:-1, :-1] = Z
            else:
                lum[..., :] = np.expand_dims(Z, axis=2)

            glstate.m_vbo_lum.bind()
            glstate.m_vbo_lum.write(0, lum, lum.nbytes)
            glstate.m_vbo_lum.release()

            dirty_bits &= ~DirtyFlag.Z

        glstate.render_cache = [origin, dirty_bits]

        widget.setViewboxClip(view)

        glstate.m_vao.bind()
        glstate.m_texture.bind()
        program.bind()

        lo, hi = self.levels
        rng = hi - lo
        if rng == 0:
            rng = 1
        OpenGLHelpers.setUniformValue(program, "u_rescale", QtGui.QVector2D(1/rng, lo))

        OpenGLHelpers.setUniformValue(program, "u_mvp", mvp)

        NULL = compat.voidptr(0) if QT_LIB.startswith("PySide") else None
        glfn.glDrawElements(GLC.GL_TRIANGLES, num_ind_mesh, GLC.GL_UNSIGNED_INT, NULL)

        glstate.m_vao.release()


class OpenGLState(QtCore.QObject):
    VERT_SRC_COMPAT = """
        attribute vec4 a_position;
        attribute float a_luminance;
        varying float v_luminance;
        uniform mat4 u_mvp;
        uniform vec2 u_rescale;
        void main() {
            v_luminance = u_rescale.x * (a_luminance - u_rescale.y);
            gl_Position = u_mvp * a_position;
        }
    """
    FRAG_SRC_COMPAT = """
        #ifdef GL_ES
        precision mediump float;
        #endif
        varying float v_luminance;
        uniform sampler2D u_texture;
        void main() {
            if (!(v_luminance == v_luminance)) discard;
            float s = clamp(v_luminance, 0.0, 1.0);
            gl_FragColor = texture2D(u_texture, vec2(s, 0));
        }
    """

    VERT_SRC = """
        in vec4 a_position;
        in float a_luminance;
        flat out float v_luminance;
        uniform mat4 u_mvp;
        uniform vec2 u_rescale;
        void main() {
            v_luminance = u_rescale.x * (a_luminance - u_rescale.y);
            gl_Position = u_mvp * a_position;
        }
    """
    FRAG_SRC = """
        #ifdef GL_ES
        precision mediump float;
        #endif
        flat in float v_luminance;
        out vec4 FragColor;
        uniform sampler2D u_texture;
        void main() {
            if (isnan(v_luminance)) discard;
            float s = clamp(v_luminance, 0.0, 1.0);
            FragColor = texture(u_texture, vec2(s, 0));
        }
    """

    def __init__(self, parent):
        super().__init__(parent)
        self.context = None
        self.render_cache = None
        self.m_vao = QtOpenGL.QOpenGLVertexArrayObject(self)
        self.m_vbo_pos = QtOpenGL.QOpenGLBuffer(QtOpenGL.QOpenGLBuffer.Type.VertexBuffer)
        self.m_vbo_lum = QtOpenGL.QOpenGLBuffer(QtOpenGL.QOpenGLBuffer.Type.VertexBuffer)
        self.m_vbo_ind = QtOpenGL.QOpenGLBuffer(QtOpenGL.QOpenGLBuffer.Type.IndexBuffer)
        self.m_texture = QtOpenGL.QOpenGLTexture(QtOpenGL.QOpenGLTexture.Target.Target2D)

    def setup(self, context):
        if self.context is context:
            return

        if self.context is not None:
            self.context.aboutToBeDestroyed.disconnect(self.cleanup)
            self.cleanup()

        self.context = context
        self.context.aboutToBeDestroyed.connect(self.cleanup)

        is_opengles = self.context.isOpenGLES()
        gl_version = self.context.format().version()
        if not is_opengles and gl_version >= (3, 1):
            moderngl = True
        elif is_opengles and gl_version >= (3, 0):
            moderngl = True
        else:
            moderngl = False

        if moderngl:
            self.flat_shading = True
            if not is_opengles:
                glsl_version = "#version 140"
            else:
                glsl_version = "#version 300 es"
            VERT_SRC = "\n".join([glsl_version, OpenGLState.VERT_SRC])
            FRAG_SRC = "\n".join([glsl_version, OpenGLState.FRAG_SRC])
        else:
            self.flat_shading = False
            VERT_SRC = OpenGLState.VERT_SRC_COMPAT
            FRAG_SRC = OpenGLState.FRAG_SRC_COMPAT

        glwidget = self.parent()
        program = glwidget.retrieveProgram("PColorMeshItem")
        if program is None:
            program = QtOpenGL.QOpenGLShaderProgram()
            if not program.addShaderFromSourceCode(QtOpenGL.QOpenGLShader.ShaderTypeBit.Vertex, VERT_SRC):
                raise RuntimeError(program.log())
            if not program.addShaderFromSourceCode(QtOpenGL.QOpenGLShader.ShaderTypeBit.Fragment, FRAG_SRC):
                raise RuntimeError(program.log())
            program.bindAttributeLocation("a_position", 0)
            program.bindAttributeLocation("a_luminance", 1)
            if not program.link():
                raise RuntimeError(program.log())
        glwidget.storeProgram("PColorMeshItem", program)

        self.m_vao.create()
        self.m_vbo_pos.create()
        self.m_vbo_lum.create()
        self.m_vbo_ind.create()


        loc_pos, loc_lum = 0, 1
        self.m_vao.bind()
        self.m_vbo_ind.bind()
        self.m_vbo_pos.bind()
        program.enableAttributeArray(loc_pos)
        program.setAttributeBuffer(loc_pos, GLC.GL_FLOAT, 0, 2)
        self.m_vbo_pos.release()
        self.m_vbo_lum.bind()
        program.enableAttributeArray(loc_lum)
        program.setAttributeBuffer(loc_lum, GLC.GL_FLOAT, 0, 1)
        self.m_vbo_lum.release()
        self.m_vao.release()
        self.m_vbo_ind.release()

    def cleanup(self):
        # this method should restore the state back to __init__
        glwidget = self.parent()
        glwidget.makeCurrent()

        for name in ['m_texture', 'm_vbo_pos', 'm_vbo_lum', 'm_vbo_ind', 'm_vao']:
            obj = getattr(self, name)
            obj.destroy()

        self.context = None
        self.render_cache = None

        glwidget.doneCurrent()

    def setTextureLut(self, lut):
        tex = self.m_texture
        if not tex.isCreated():
            tex.setFormat(tex.TextureFormat.RGBAFormat)
            tex.setSize(256, 1)
            tex.allocateStorage()
            tex.setMinMagFilters(tex.Filter.Nearest, tex.Filter.Nearest)
            tex.setWrapMode(tex.WrapMode.ClampToEdge)
        tex.setData(tex.PixelFormat.RGBA, tex.PixelType.UInt8, lut)

    def dataChange(self, dirtyFlags : DirtyFlag):
        if self.render_cache is None:
            return

        if DirtyFlag.XY in dirtyFlags:
            self.render_cache[0] = None
            self.render_cache[1] |= DirtyFlag.XY

        if DirtyFlag.Z in dirtyFlags:
            self.render_cache[1] |= DirtyFlag.Z

        if DirtyFlag.DIM in dirtyFlags:
            self.render_cache[1] |= DirtyFlag.DIM

        if DirtyFlag.LUT in dirtyFlags:
            self.render_cache[1] |= DirtyFlag.LUT



import numpy as np

from .. import colormap
from .. import functions as fn
from ..Qt import QtCore, QtGui, QtWidgets
from .ColorMapMenu import ColorMapMenu

__all__ = ['ColorMapButton']


class ColorMapDisplayMixin:
    # Note that this Mixin attempts to be safe for inheritance
    # on either the lhs or rhs. To this end:
    # 1) the __init__ is safe to be called more than once
    # 2) it does not override any main class virtual methods

    def __init__(self, *, orientation):
        self.horizontal = orientation == 'horizontal'
        self._menu = None
        self._setColorMap(None)

    def setMaximumThickness(self, val):
        # calls main class methods
        Thickness = 'Height' if self.horizontal else 'Width'
        getattr(self, f'setMaximum{Thickness}')(val)

    def _setColorMap(self, cmap):
        if isinstance(cmap, str):
            try:
                cmap = colormap.get(cmap)
            except FileNotFoundError:
                cmap = None
        if cmap is None:
            cmap = colormap.ColorMap(None, [0.0, 1.0])

        self._cmap = cmap
        self._image = None

    def setColorMap(self, cmap):
        # calls main class methods
        self._setColorMap(cmap)
        self.colorMapChanged()

    def colorMap(self):
        return self._cmap

    def getImage(self):
        if self._image is None:
            lut = self._cmap.getLookupTable(nPts=256, alpha=True)
            lut = np.expand_dims(lut, axis=0 if self.horizontal else 1)
            qimg = fn.ndarray_to_qimage(lut, QtGui.QImage.Format.Format_RGBA8888)
            self._image = qimg if self.horizontal else qimg.mirrored()
        return self._image

    def getMenu(self):
        if self._menu is None:
            self._menu = ColorMapMenu(showColorMapSubMenus=True)
            self._menu.sigColorMapTriggered.connect(self.setColorMap)
        return self._menu

    def paintColorMap(self, painter, rect):
        painter.save()
        image = self.getImage()
        painter.drawImage(rect, image)

        if not self.horizontal:
            painter.translate(rect.center())
            painter.rotate(-90)
            painter.translate(-rect.center())

        text = self.colorMap().name
        wpen = QtGui.QPen(QtCore.Qt.GlobalColor.white)
        bpen = QtGui.QPen(QtCore.Qt.GlobalColor.black)
        # get an estimate of the lightness of the colormap
        # from its center element
        lightness = image.pixelColor(image.rect().center()).lightnessF()
        pen = bpen if lightness >= 0.55 else wpen

        AF = QtCore.Qt.AlignmentFlag
        trect = painter.boundingRect(rect, AF.AlignCenter, text)
        # draw the foreground text
        painter.setPen(pen)
        painter.drawText(trect, text)

        painter.restore()


class ColorMapButton(ColorMapDisplayMixin, QtWidgets.QWidget):
    sigColorMapChanged = QtCore.Signal(object)

    def __init__(self):
        QtWidgets.QWidget.__init__(self)
        ColorMapDisplayMixin.__init__(self, orientation='horizontal')

    def colorMapChanged(self):
        cmap = self.colorMap()
        self.sigColorMapChanged.emit(cmap)
        self.update()

    def paintEvent(self, evt):
        painter = QtGui.QPainter(self)
        self.paintColorMap(painter, self.contentsRect())
        painter.end()

    def mouseReleaseEvent(self, evt):
        if evt.button() != QtCore.Qt.MouseButton.LeftButton:
            return

        # position the menu below the widget
        pos = self.mapToGlobal(self.pos())
        pos.setY(pos.y() + self.height())
        self.getMenu().popup(pos)



#!/usr/bin/python
"""
Update a simple plot as rapidly as possible to measure speed.
"""

import argparse
import itertools

import numpy as np
from utils import FrameCounter

import pyqtgraph as pg
import pyqtgraph.functions as fn
import pyqtgraph.parametertree as ptree
from pyqtgraph.Qt import QtCore, QtGui, QtWidgets

# defaults here result in the same configuration as the original PlotSpeedTest
parser = argparse.ArgumentParser()
parser.add_argument('--noise', dest='noise', action='store_true')
parser.add_argument('--no-noise', dest='noise', action='store_false')
parser.set_defaults(noise=True)
parser.add_argument('--nsamples', default=5000, type=int)
parser.add_argument('--frames', default=50, type=int)
parser.add_argument('--fsample', default=1000, type=float)
parser.add_argument('--frequency', default=0, type=float)
parser.add_argument('--amplitude', default=5, type=float)
parser.add_argument('--opengl', dest='use_opengl', action='store_true')
parser.add_argument('--no-opengl', dest='use_opengl', action='store_false')
parser.set_defaults(use_opengl=None)
parser.add_argument('--allow-opengl-toggle', action='store_true',
    help="""Allow on-the-fly change of OpenGL setting. This may cause unwanted side effects.
    """)
parser.add_argument('--iterations', default=float('inf'), type=float,
    help="Number of iterations to run before exiting"
)
args = parser.parse_args()

if args.use_opengl is not None:
    pg.setConfigOption('useOpenGL', args.use_opengl)
use_opengl = pg.getConfigOption('useOpenGL')

# don't limit frame rate to vsync
sfmt = QtGui.QSurfaceFormat()
sfmt.setSwapInterval(0)
QtGui.QSurfaceFormat.setDefaultFormat(sfmt)


class MonkeyCurveItem(pg.PlotCurveItem):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.monkey_mode = ''

    def setMethod(self, value):
        self.monkey_mode = value

    def paint(self, painter, opt, widget):
        if self.monkey_mode not in ['drawPolyline']:
            return super().paint(painter, opt, widget)

        painter.setRenderHint(painter.RenderHint.Antialiasing, self.opts['antialias'])
        painter.setPen(pg.mkPen(self.opts['pen']))

        if self.monkey_mode == 'drawPolyline':
            painter.drawPolyline(fn.arrayToQPolygonF(self.xData, self.yData))

app = pg.mkQApp("Plot Speed Test")

default_pen = pg.mkPen()

params = ptree.Parameter.create(name='Parameters', type='group')
pt = ptree.ParameterTree(showHeader=False)
pt.setParameters(params)
pw = pg.PlotWidget()
splitter = QtWidgets.QSplitter()
splitter.addWidget(pt)
splitter.addWidget(pw)
splitter.show()

interactor = ptree.Interactor(
    parent=params, nest=False, runOptions=ptree.RunOptions.ON_CHANGED
)

pw.setWindowTitle('pyqtgraph example: PlotSpeedTest')
pw.setLabel('bottom', 'Index', units='B')
curve = MonkeyCurveItem(pen=default_pen, brush='b')
pw.addItem(curve)
iterations_counter = itertools.count()

@interactor.decorate(
    nest=True,
    nsamples={'limits': [0, None]},
    frames={'limits': [1, None]},
    fsample={'units': 'Hz'},
    frequency={'units': 'Hz'}
)
def makeData(
    noise=args.noise,
    nsamples=args.nsamples,
    frames=args.frames,
    fsample=args.fsample,
    frequency=args.frequency,
    amplitude=args.amplitude,
):
    global data, connect_array, ptr
    ttt = np.arange(frames * nsamples, dtype=np.float64) / fsample
    data = amplitude*np.sin(2*np.pi*frequency*ttt).reshape((frames, nsamples))
    if noise:
        data += np.random.normal(size=data.shape)
    connect_array = np.ones(data.shape[-1], dtype=bool)
    ptr = 0
    pw.setRange(QtCore.QRectF(0, -10, nsamples, 20))

params.child('makeData').setOpts(title='Plot Options')


@interactor.decorate(
    connect={'type': 'list', 'limits': ['all', 'pairs', 'finite', 'array']}
)
def update(
    antialias=pg.getConfigOption('antialias'),
    connect='all',
    skipFiniteCheck=False
):
    global ptr

    if next(iterations_counter) > args.iterations:
        # cleanly close down benchmark
        timer.stop()
        app.quit()
        return None

    if connect == 'array':
        connect = connect_array

    curve.setData(
        data[ptr],
        antialias=antialias,
        connect=connect,
        skipFiniteCheck=skipFiniteCheck
    )
    ptr = (ptr + 1) % data.shape[0]
    framecnt.update()


@interactor.decorate(
    useOpenGL={'readonly': not args.allow_opengl_toggle},
    plotMethod={'limits': ['pyqtgraph', 'drawPolyline'], 'type': 'list'},
    curvePen={'type': 'pen'}
)
def updateOptions(
    curvePen=pg.mkPen(),
    plotMethod='pyqtgraph',
    fillLevel=False,
    enableExperimental=False,
    useOpenGL=use_opengl,
):
    pg.setConfigOption('enableExperimental', enableExperimental)
    pw.useOpenGL(useOpenGL)
    curve.setPen(curvePen)
    curve.setFillLevel(0.0 if fillLevel else None)
    curve.setMethod(plotMethod)


makeData()

timer = QtCore.QTimer()
timer.timeout.connect(update)
timer.start(0)

framecnt = FrameCounter()
framecnt.sigFpsUpdate.connect(lambda fps: pw.setTitle(f'{fps:.1f} fps'))

if __name__ == '__main__':
    # Splitter by default gives too small of a width to the parameter tree,
    # so fix that right before the event loop
    pt.setMinimumSize(225,0)
    pg.exec()



import weakref
from math import ceil, floor, frexp, isfinite, log10, sqrt

import numpy as np

from .. import debug as debug
from .. import functions as fn
from .. import getConfigOption
from ..Point import Point
from ..Qt import QtCore, QtGui, QtWidgets
from .GraphicsWidget import GraphicsWidget

__all__ = ['AxisItem']


class AxisItem(GraphicsWidget):
    """
    GraphicsItem showing a single plot axis with ticks, values, and label.
    
    Can be configured to fit on any side of a plot, automatically synchronize its
    displayed scale with ViewBox items. Ticks can be extended to draw a grid.
    
    If maxTickLength is negative, ticks point into the plot.

    Parameters
    ----------
    orientation : {'left', 'right', 'top', 'bottom'}
        The side of the plot the axis is attached to.
    pen : QPen or None
        Pen used when drawing axis and (by default) ticks.
    textPen : QPen or None
        Pen used when drawing tick labels.
    tickPen : QPen or None
        Pen used when drawing ticks.
    linkView : ViewBox or None
        Causes the range of values displayed in the axis to be linked to the visible
        range of a ViewBox.
    parent : QtWidgets.QGraphicsItem or None
        Parent Qt object to set to. End users are not expected to set, pyqtgraph should
        set correctly on its own.
    maxTickLength : int
        Maximum length of ticks to draw in pixels. Negative values draw into the
        plot, positive values draw outward.  Default -5.
    showValues : bool
        Whether to display values adjacent to ticks. Default true.
    **args
        All additional keyword arguments are passed to :func:`setLabel`.
    """
    def __init__(
            self,
            orientation: str,
            pen=None,
            textPen=None,
            tickPen = None,
            linkView=None,
            parent=None,
            maxTickLength=-5,
            showValues=True,
            **args,
    ):
        super().__init__(parent)
        self.label = QtWidgets.QGraphicsTextItem(self)
        self.picture = None
        self.orientation = orientation

        if orientation in {'left', 'right'}:
            self.label.setRotation(-90)
            # allow labels on vertical axis to extend above and below the length of the axis
            hide_overlapping_labels = False
        elif orientation in {'top', 'bottom'}:
            # stop labels on horizontal axis from overlapping so vertical axis labels have room
            hide_overlapping_labels = True
        else:
            raise ValueError(
                "Orientation argument must be one of 'left', 'right', 'top', or 'bottom'."
            )
        self.style = {
            'tickTextOffset': [5, 2],  ## (horizontal, vertical) spacing between text and axis
            'tickTextWidth': 30,  ## space reserved for tick text
            'tickTextHeight': 18,
            'autoExpandTextSpace': True,  ## automatically expand text space if needed
            'autoReduceTextSpace': True,
            'hideOverlappingLabels': hide_overlapping_labels,
            'tickFont': None,
            'stopAxisAtTick': (False, False),  ## whether axis is drawn to edge of box or to last tick
            'textFillLimits': [  ## how much of the axis to fill up with tick text, maximally.
                (0, 0.8),    ## never fill more than 80% of the axis
                (2, 0.6),    ## If we already have 2 ticks with text, fill no more than 60% of the axis
                (4, 0.4),    ## If we already have 4 ticks with text, fill no more than 40% of the axis
                (6, 0.2),    ## If we already have 6 ticks with text, fill no more than 20% of the axis
            ],
            'showValues': showValues,
            'tickLength': maxTickLength,
            'maxTickLevel': 2,
            'maxTextLevel': 2,
            'tickAlpha': None,  ## If not none, use this alpha for all ticks.
        }

        self.textWidth = 30  ## Keeps track of maximum width / height of tick text
        self.textHeight = 18

        # If the user specifies a width / height, remember that setting
        # indefinitely.
        self.fixedWidth = None
        self.fixedHeight = None

        self.logMode = False

        self._tickDensity = 1.0   # used to adjust scale the number of automatically generated ticks
        self._tickLevels  = None  # used to override the automatic ticking system with explicit ticks
        self._tickSpacing = None  # used to override default tickSpacing method
        self.scale = 1.0
        self.autoSIPrefix = True
        self.autoSIPrefixScale = 1.0

        self.labelText = ""
        self.labelUnits = ""
        self.labelUnitPrefix = ""
        self.unitPower = 1
        self.labelStyle = {}
        self._siPrefixEnableRanges = None
        self.setRange(0, 1)
        self.setLabel(**args)
        self.showLabel(False)


        if pen is None:
            self.setPen()
        else:
            self.setPen(pen)

        if textPen is None:
            self.setTextPen()
        else:
            self.setTextPen(textPen)

        if tickPen is None:
            self.setTickPen()
        else:
            self.setTickPen(tickPen)

        self._linkedView = None
        if linkView is not None:
            self._linkToView_internal(linkView)

        self.grid = False

        #self.setCacheMode(self.DeviceCoordinateCache)

    def setStyle(self, **kwargs):
        """
        Set various style options.

        Parameters
        ----------
        **kwargs : dict, optional
            Here are a list of supported arguments.

            ===================== ======================================================
            Property              Description
            ===================== ======================================================
            tickLength            ``int``
                                  The maximum length of ticks in pixels. Positive values
                                  point toward the text; negative values point away.

            tickTextOffset        ``int`` 
                                  Reserved spacing between text and axis in pixels.

            tickTextWidth         ``int``
                                  Horizontal space reserved for tick text in pixels.

            tickTextHeight        ``int``
                                  Vertical space reserved for tick text in pixels.

            autoExpandTextSpace   ``bool``
                                  Automatically expand text space if the tick strings
                                  become too long.

            autoReduceTextSpace   ``bool``
                                  Automatically shrink the axis if necessary.

            hideOverlappingLabels ``bool`` or ``int``

                                  - ``True``  (default for horizontal axis): Hide tick
                                    labels which extend beyond the AxisItem's geometry
                                    rectangle.
                                  - ``False`` (default for vertical axis): Labels may be
                                    drawn extending beyond the extent of the axis.
                                  - ``int`` sets the tolerance limit for how many pixels
                                    a label is allowed to extend beyond the axis.
                                    Defaults to 15 for
                                    ``hideOverlappingLabels = False``.

            tickFont              :class:`QFont` or ``None``
                                  Determines the font used for tick values. Use None for
                                  the default font.
            
            stopAxisAtTick        tuple of ``bool, bool`` 
                                  The first element represents the horizontal axis, the
                                  second element represents the vertical axis.

                                  - ``True`` - The axis line is drawn only as far as the
                                    last tick.
                                  - ``False`` - The line is drawn to the edge of the
                                    :class:`~pyqtgraph.AxisItem` boundary.

            textFillLimits        list of ``(int, float)``
                                  This structure determines how the AxisItem decides how
                                  many ticks should have text appear next to them.
                                  The first value corresponds to the tick number.  The
                                  second value corresponds to the fill percentage. Each
                                  tuple in the list specifies what fraction of the axis
                                  length may be occupied by text, given the number of
                                  ticks that already have text displayed.
                                  
                                  For example ::

                                    [
                                        # Never fill more than 80% of the axis
                                        (0, 0.8),
                                        # If we already have 2 ticks with text, fill no
                                        # more than 60% of the axis
                                        (2, 0.6), 
                                        # If we already have 4 ticks with text, fill no
                                        # more than 40% of the axis
                                        (4, 0.4), 
                                        # If we already have 6 ticks with text, fill no
                                        # more than 20% of the axis
                                        (6, 0.2)
                                    ]
                                                
            showValues            ``bool``
                                  indicates whether text is displayed adjacent to ticks.
            
            tickAlpha             ``float``, ``int`` or ``None`` 
                                  If ``None``, pyqtgraph will draw the ticks with the
                                  alpha it deems appropriate. Otherwise, the alpha will
                                  be fixed at the value passed. With ``int``, accepted
                                  values are [0..255]. With value of type ``float``,
                                  accepted values are from [0..1].

            maxTickLevel          ``int``
                                  default: 2

                                  Tick (and grid line) density level.

                                  - 0: Show major ticks only
                                  - 1: Show major ticks and one level of minor ticks
                                  - 2: Show major ticks and two levels of minor ticks
                                    (higher CPU usage)
            ===================== ======================================================

        Raises
        ------
        NameError
            Raised when the name of a keyword argument is not recognized.
        TypeError
            Raised when a value for a keyword argument is of the wrong type.
        """
        for kwd, value in kwargs.items():
            if kwd not in self.style:
                raise NameError(f"{kwd} is not a valid style argument.")

            if (
                kwd in (
                    'tickLength',
                    'tickTextOffset',
                    'tickTextWidth',
                    'tickTextHeight'
                ) and 
                not isinstance(value, int)
            ):
                raise TypeError(f"Argument '{kwd}' must be int")

            if kwd == 'tickTextOffset':
                if self.orientation in ('left', 'right'):
                    self.style['tickTextOffset'][0] = value
                else:
                    self.style['tickTextOffset'][1] = value
            elif kwd == 'stopAxisAtTick':
                if len(value) != 2 or not all(isinstance(val, bool) for val in value):
                    raise TypeError(
                        "Argument 'stopAxisAtTick' must have type (bool, bool)"
                    )
                self.style[kwd] = value
            else:
                self.style[kwd] = value

        self.picture = None
        self._adjustSize()
        self.update()

    def close(self):
        self.scene().removeItem(self.label)
        self.label = None
        self.scene().removeItem(self)

    def setGrid(self, grid: int | float | bool):
        """
        Set the alpha value for the grid, or ``False`` to disable.

        When grid lines are enabled, the axis tick lines are extended to cover the
        extent of the linked ViewBox, if any.

        Parameters
        ----------
        grid : bool or int or float
            Alpha value to apply to :class:`~pyqtgraph.GridItem`.
            
            - ``False`` - Disable the grid.
            - ``int`` - Values between [0, 255] to set the alpha of the grid to.
            - ``float`` - Values between [0..1] to set the alpha of the grid to.
        """
        if isinstance(grid, float):
            grid = int(grid * 255)
            grid = min(grid, 255)
            grid = max(grid, 0)
        self.grid = grid
        self.picture = None
        self.prepareGeometryChange()
        self.update()

    def setLogMode(
        self,
        *args: tuple[bool] | tuple[bool, bool] | None,
        **kwargs: dict[str, bool] | None
    ):
        """
        Set log scaling for x and / or y axes.

        If two positional arguments are provided, the first will set log scaling
        for the x axis and the second for the y axis. If a single positional
        argument is provided, it will set the log scaling along the direction of
        the AxisItem. Alternatively, x and y can be passed as keyword arguments.

        If an axis is set to log scale, ticks are displayed on a logarithmic scale and
        values are adjusted accordingly. The linked ViewBox will be informed of the
        change.

        Parameters
        ----------
        *args : tuple of bool
            If length 1, sets log mode regardless of orientation.  If length 2, the
            first element toggles log mode for x-axis, and the second element toggles
            log mode for the y-axis.
        **kwargs : dict
            Pass a dictionary with keys `x` and `y`, where the values are ``bool`` to
            set the log mode for the respective `x` or `y` axis.  Trying to set the `y`
            axis log mode while this axis item is horizontal (or vice versa) will be
            ignored.

        See Also
        --------
        :meth:`~pyqtgraph.PlotItem.setLogMode`
            The method called to shift the values of the data.
        """
        if len(args) == 1:
            self.logMode = args[0]
        else:
            if len(args) == 2:
                x, y = args
            else:
                x = kwargs.get('x')
                y = kwargs.get('y')

            if x is not None and self.orientation in ('top', 'bottom'):
                self.logMode = x
            if y is not None and self.orientation in ('left', 'right'):
                self.logMode = y

        # inform the linked views of the change
        if self._linkedView is not None:
            if self.orientation in ('top', 'bottom'):
                self._linkedView().setLogMode('x', self.logMode)
            elif self.orientation in ('left', 'right'):
                self._linkedView().setLogMode('y', self.logMode)

        self.picture = None
        self.update()

    def setTickFont(self, font: QtGui.QFont | None):
        """
        Set the font used for tick values.
        
        Parameters
        ----------
        font : QtGui.QFont or None
            The font to use for the tick values. Set to ``None`` for the default font.
        """
        self.style['tickFont'] = font
        self.picture = None
        self.prepareGeometryChange()
        # Need to re-allocate space depending on font size?
        self.update()

    def resizeEvent(self, ev=None):
        # Set the position of the label
        nudge = 5
        # self.label is set to None on close, but resize events can still occur.
        if self.label is None:
            self.picture = None
            return

        br = self.label.boundingRect()
        p = QtCore.QPointF(0, 0)
        if self.orientation == 'left':
            p.setY(int(self.size().height()/2 + br.width()/2))
            p.setX(-nudge)
        elif self.orientation == 'right':
            p.setY(int(self.size().height()/2 + br.width()/2))
            p.setX(int(self.size().width()-br.height()+nudge))
        elif self.orientation == 'top':
            p.setY(-nudge)
            p.setX(int(self.size().width()/2. - br.width()/2.))
        elif self.orientation == 'bottom':
            p.setX(int(self.size().width()/2. - br.width()/2.))
            p.setY(int(self.size().height()-br.height()+nudge))
        self.label.setPos(p)
        self.picture = None

    def showLabel(self, show: bool=True):
        """
        Show or hide the label text for this axis.

        Parameters
        ----------
        show : bool, optional
            Show the label text, by default True.
        """
        self.label.setVisible(show)
        if self.orientation in ['left', 'right']:
            self._updateWidth()
        else:
            self._updateHeight()
        if self.autoSIPrefix:
            self.updateAutoSIPrefix()

    def setLabel(
        self,
        text: str | None=None,
        units: str | None=None,
        unitPrefix: str | None=None,
        siPrefixEnableRanges: tuple[tuple[float, float], ...] | None=None,
        unitPower: int | float=1,
        **kwargs
    ):
        """
        Set the text displayed adjacent to the axis.

        Parameters
        ----------
        text : str
            The text (excluding units) to display on the label for this axis.
        units : str
            The units for this axis. Units should generally be given without any scaling
            prefix (eg, 'V' instead of 'mV'). The scaling prefix will be automatically
            prepended based on the range of data displayed.
        unitPrefix : str
            An extra prefix to prepend to the units.
        siPrefixEnableRanges : tuple of tuple of float, float, Optional
            The ranges in which automatic SI prefix scaling is enabled. Defaults to
            everywhere, unless units is empty, in which case it defaults to
            ``((0., 1.), (1e9, inf))``.
        unitPower : int or float, optional
            The power to which the units are raised. For example, if units='m²', the
            unitPower should be 2. This ensures correct scaling when using SI prefixes.
            Supports positive, negative and non-integral powers.  Default is 1.
            Note: The power only affects the scaling, not the units themselves. For
            example, with units='m' and unitPower=2, the displayed units will still be 'm'.
        **kwargs
            All extra keyword arguments become CSS style options for the ``<span>`` tag
            which will surround the axis label and units. Note that CSS attributes are
            not always valid python arguments. Examples: ``color='#FFF'``,
            ``**{'font-size': '14pt'}``.

        Notes
        -----
        The final text generated for the label will usually take the form::

            <span style="...args...">{text} (prefix{units})</span>
        """
        self.labelText = text or ""
        self.labelUnits = units or ""
        self.labelUnitPrefix = unitPrefix or ""
        self.unitPower = unitPower
        if kwargs:
            self.labelStyle = kwargs
        self.setSIPrefixEnableRanges(siPrefixEnableRanges)
        # Account empty string and `None` for units and text
        visible = bool(text or units)
        self.showLabel(visible)
        self._updateLabel()

    def setSIPrefixEnableRanges(self, ranges=None):
        """
        Set the ranges in which automatic SI prefix scaling is enabled.

        This function allows you to define specific ranges where SI prefixes will be
        used. By default, SI prefix scaling is enabled everywhere, unless units are
        empty, in which case it defaults to ``((0., 1.), (1e9, inf))``.

        Parameters
        ----------
        ranges : tuple of tuple of float, float, optional
            A tuple of ranges where SI prefix scaling is enabled. Each range is a tuple
            containing two floats representing the start and end of the range.
        """
        self._siPrefixEnableRanges = ranges

    def getSIPrefixEnableRanges(self):
        """
        Get the ranges in which automatic SI prefix scaling is enabled.

        Returns
        -------
        tuple of tuple of float, float
            A tuple of ranges where SI prefix scaling is enabled. Each range is a tuple
            containing two floats representing the start and end of the range. If no
            custom ranges are set, then the default ranges are returned. The default
            ranges are ``((0., 1.), (1e9, inf))`` if units are empty, and 
            ``((0., inf))`` otherwise.
        """
        if self._siPrefixEnableRanges is not None:
            return self._siPrefixEnableRanges
        elif self.labelUnits == '':
            return (0., 1.), (1e9, float('inf'))
        else:
            return ((0., float('inf')),)

    def _updateLabel(self):
        self.label.setHtml(self.labelString())
        self._adjustSize()
        self.picture = None
        self.update()

    def labelString(self) -> str:
        """
        Generate the label string based on current label, units, and prefix.

        Returns
        -------
        str
            The complete label string, including units and any prefixes.
        """
        if self.labelUnits == '':
            if not self.autoSIPrefix or self.autoSIPrefixScale == 1.0:
                units = ''
            else:
                units = f'(x{1.0 / self.autoSIPrefixScale:g})'
        else:
            units = f'({self.labelUnitPrefix}{self.labelUnits})'

        s = f'{self.labelText} {units}'

        style = ';'.join([f'{k}: {self.labelStyle[k]}' for k in self.labelStyle])

        return f"<span style='{style}'>{s}</span>"

    def _updateMaxTextSize(self, x: int):
        ## Informs that the maximum tick size orthogonal to the axis has
        ## changed; we use this to decide whether the item needs to be resized
        ## to accommodate.
        if self.orientation in ['left', 'right']:
            if self.style["autoReduceTextSpace"]:
                if x > self.textWidth or x < self.textWidth - 10:
                    self.textWidth = x
            else:
                mx = max(self.textWidth, x)
                if mx > self.textWidth or mx < self.textWidth - 10:
                    self.textWidth = mx
            if self.style['autoExpandTextSpace']:
                self._updateWidth()

        else:
            if self.style['autoReduceTextSpace']:
                if x > self.textHeight or x < self.textHeight - 10:
                    self.textHeight = x
            else:
                mx = max(self.textHeight, x)
                if mx > self.textHeight or mx < self.textHeight - 10:
                    self.textHeight = mx
            if self.style['autoExpandTextSpace']:
                self._updateHeight()

    def _adjustSize(self):
        if self.orientation in ['left', 'right']:
            self._updateWidth()
        else:
            self._updateHeight()

    def setHeight(self, h: int | None=None):
        """
        Set the height of this axis reserved for ticks and tick labels.

        The height of the axis label is automatically added.

        Parameters
        ----------
        h : int or None, optional
            If ``None``, then the value will be determined automatically based on the
            size of the tick text, by default None.
        """
        self.fixedHeight = h
        self._updateHeight()

    def _updateHeight(self):
        if not self.isVisible():
            h = 0
        elif self.fixedHeight is None:
            if not self.style['showValues']:
                h = 0
            elif self.style['autoExpandTextSpace']:
                h = self.textHeight
            else:
                h = self.style['tickTextHeight']
            h += self.style['tickTextOffset'][1] if self.style['showValues'] else 0
            h += max(0, self.style['tickLength'])
            if self.label.isVisible():
                h += self.label.boundingRect().height() * 0.8
        else:
            h = self.fixedHeight

        self.setMaximumHeight(h)
        self.setMinimumHeight(h)
        self.picture = None

    def setWidth(self, w: int | None=None):
        """
        Set the width of this axis reserved for ticks and tick labels.

        The width of the axis label is automatically added.

        Parameters
        ----------
        w : int or None, optional
            If ``None``, then the value will be determined automatically based on the
            size of the tick text, by default None.
        """        
        self.fixedWidth = w
        self._updateWidth()

    def _updateWidth(self):
        if not self.isVisible():
            w = 0
        elif self.fixedWidth is None:
            if not self.style['showValues']:
                w = 0
            elif self.style['autoExpandTextSpace']:
                w = self.textWidth
            else:
                w = self.style['tickTextWidth']
            w += self.style['tickTextOffset'][0] if self.style['showValues'] else 0
            w += max(0, self.style['tickLength'])
            if self.label.isVisible():
                w += self.label.boundingRect().height() * 0.8  ## bounding rect is usually an overestimate
        else:
            w = self.fixedWidth

        self.setMaximumWidth(w)
        self.setMinimumWidth(w)
        self.picture = None

    def pen(self) -> QtGui.QPen:
        """
        Get the pen used for drawing text, axes, ticks, and grid lines.

        If no custom pen has been set, this method will return a pen with the
        default foreground color.

        Returns
        -------
        QPen
            The pen used to draw text, axes, ticks, and grid lines.
        """
        if self._pen is None:
            return fn.mkPen(getConfigOption('foreground'))
        return fn.mkPen(self._pen)

    def setPen(self, *args, **kwargs):
        """
        Set the pen used for drawing text, axes, ticks, and grid lines.
        
        If no arguments given, the default foreground color will be used.

        Parameters
        ----------
        *args : tuple
            Arguments relayed to :func:`~pyqtgraph.mkPen`.
        **kwargs : dict
            Arguments relayed to `:func:`~pyqtgraph.mkPen`.

        See Also
        --------
        :func:`setConfigOption <pyqtgraph.setConfigOption>`
            Option to change the default foreground color.
        """        
        self.picture = None
        if args or kwargs:
            self._pen = fn.mkPen(*args, **kwargs)
        else:
            self._pen = fn.mkPen(getConfigOption('foreground'))
        self.labelStyle['color'] = self._pen.color().name() #   #RRGGBB
        self._updateLabel()

    def textPen(self) -> QtGui.QPen:
        """
        Get the pen used for drawing text.

        If no custom text pen has been set, this method will return a pen with the
        default foreground color.

        Returns
        -------
        QPen
            The pen used to draw text.
        """
        if self._textPen is None:
            return fn.mkPen(getConfigOption('foreground'))
        return fn.mkPen(self._textPen)

    def setTextPen(self, *args, **kwargs):
        """
        Set the pen used for drawing text.

        If no arguments given, the default foreground color will be used.
        
        Parameters
        ----------
        *args : tuple
            Arguments relayed to :func:`~pyqtgraph.mkPen`.
        **kwargs : dict
            Arguments relayed to `:func:`~pyqtgraph.mkPen`.

        See Also
        --------
        :func:`setConfigOption <pyqtgraph.setConfigOption>`
            Option to change the default foreground color.
        """     
        self.picture = None
        if args or kwargs:
            self._textPen = fn.mkPen(*args, **kwargs)
        else:
            self._textPen = fn.mkPen(getConfigOption('foreground'))
        self.labelStyle['color'] = self._textPen.color().name() #   #RRGGBB
        self._updateLabel()

    def tickPen(self) -> QtGui.QPen:
        """
        Get the pen used for drawing ticks.

        If no custom tick pen has been set, this method will return the axis's
        main pen.

        Returns
        -------
        QPen
            The pen used to draw tick marks.
        """
        return self.pen() if self._tickPen is None else fn.mkPen(self._tickPen)

    def setTickPen(self, *args, **kwargs):
        """
        Set the pen used for drawing ticks.

        If no arguments given, the default foreground color will be used.
        
        Parameters
        ----------
        *args : tuple
            Arguments relayed to :func:`~pyqtgraph.mkPen`.
        **kwargs : dict
            Arguments relayed to `:func:`~pyqtgraph.mkPen`.

        See Also
        --------
        :func:`setConfigOption <pyqtgraph.setConfigOption>`
            Option to change the default foreground color.
        """   
        self.picture = None
        self._tickPen = fn.mkPen(*args, **kwargs) if args or kwargs else None
        self._updateLabel()

    def setScale(self, scale=1.0):
        """
        Set the value scaling for this axis.

        Setting this value causes the axis to draw ticks and tick labels as if the view
        coordinate system were scaled.

        Parameters
        ----------
        scale : float, optional
            Value to scale the drawing of ticks and tick labels as if the view
            coordinate system was scaled, by default 1.0.
        """
        if scale != self.scale:
            self.scale = scale
            self._updateLabel()

    def enableAutoSIPrefix(self, enable=True):
        """
        Enable (or disable) automatic SI prefix scaling on this axis.

        When enabled, this feature automatically determines the best SI prefix
        to prepend to the label units, while ensuring that axis values are scaled
        accordingly.

        For example, if the axis spans values from -0.1 to 0.1 and has units set
        to 'V' then the axis would display values -100 to 100
        and the units would appear as 'mV'

        This feature is enabled by default, and is only available when a suffix
        (unit string) is provided to display on the label.

        Parameters
        ----------
        enable : bool, optional
            Enable Auto SI prefix, by default True.
        """

        self.autoSIPrefix = enable
        self.updateAutoSIPrefix()

    def updateAutoSIPrefix(self):
        scale = 1.0
        prefix = ''
        if self.label.isVisible():
            _range = 10**np.array(self.range) if self.logMode else self.range
            scaling_value = max(abs(_range[0]), abs(_range[1])) * self.scale
            if any(low <= scaling_value <= high for low, high in self.getSIPrefixEnableRanges()):
                (scale, prefix) = fn.siScale(scaling_value, power=self.unitPower)

        self.autoSIPrefixScale = scale
        self.labelUnitPrefix = prefix
        self._updateLabel()

    def setRange(self, mn: float, mx: float):
        """
        Set the range of values displayed by the axis.

        Usually this is handled automatically by linking the axis to a ViewBox with
        :func:`linkToView <pyqtgraph.AxisItem.linkToView>`.

        Parameters
        ----------
        mn : float
            Bottom value to set the range to.
        mx : float
            Top value to set the range to.

        Raises
        ------
        ValueError
            When non-finite values are passed.
        """

        if not isfinite(mn) or not isfinite(mx):
            raise ValueError(f"Not setting range to [{mn}, {mx}]")
        self.range = [mn, mx]
        if self.autoSIPrefix:
            # XXX: Will already update once!
            self.updateAutoSIPrefix()
        else:
            self.picture = None
            self.update()

    def linkedView(self):
        """
        Return the ViewBox linked to this axis.

        Returns
        -------
        ViewBox
            The linked ViewBox, or ``None`` if there is no ViewBox linked.
        """
        return None if self._linkedView is None else self._linkedView()

    def _linkToView_internal(self, view):
        # We need this code to be available without override,
        # even though DateAxisItem overrides the user-side linkToView method
        self.unlinkFromView()

        self._linkedView = weakref.ref(view)
        if self.orientation in ['right', 'left']:
            view.sigYRangeChanged.connect(self.linkedViewChanged)
        else:
            view.sigXRangeChanged.connect(self.linkedViewChanged)
        view.sigResized.connect(self.linkedViewChanged)

    def linkToView(self, view):
        """
        Link to a ViewBox, causing its displayed range to match the view range.

        This is usually called automatically by the ViewBox.

        Parameters
        ----------
        view : ViewBox
            The view to link to.
        """
        self._linkToView_internal(view)

    def unlinkFromView(self):
        """
        Unlink this axis from its linked ViewBox.
        """
        oldView = self.linkedView()
        self._linkedView = None
        if oldView is not None:
            oldView.sigResized.disconnect(self.linkedViewChanged)
            if self.orientation in ['right', 'left']:
                oldView.sigYRangeChanged.disconnect(self.linkedViewChanged)
            else:
                oldView.sigXRangeChanged.disconnect(self.linkedViewChanged)

    @QtCore.Slot(object)
    @QtCore.Slot(object, object)
    def linkedViewChanged(self, view, newRange=None):
        """
        Call when the linked view range has changed.

        Parameters
        ----------
        view : ViewBox
            The view whose range has changed.
        newRange : tuple of float, float, optional
            The new range of the view, by default None.
        """
        if self.orientation in ['right', 'left']:
            if newRange is None:
                newRange = view.viewRange()[1]
            if view.yInverted():
                self.setRange(*newRange[::-1])
            else:
                self.setRange(*newRange)
        else:
            if newRange is None:
                newRange = view.viewRange()[0]
            if view.xInverted():
                self.setRange(*newRange[::-1])
            else:
                self.setRange(*newRange)

    def boundingRect(self):
        m = 0
        hide_overlapping_labels = self.style['hideOverlappingLabels']
        if hide_overlapping_labels is True:
            pass # skip further checks
        elif hide_overlapping_labels is False:
            m = 15
        else:
            try:
                m = int( self.style['hideOverlappingLabels'] )
            except ValueError: pass # ignore any non-numeric value

        linkedView = self.linkedView()
        if linkedView is not None and self.grid is not False:
            return (
                self.mapRectFromParent(self.geometry()) | 
                linkedView.mapRectToItem(self, linkedView.boundingRect())
            )
        rect = self.mapRectFromParent(self.geometry())
        ## extend rect if ticks go in negative direction
        ## also extend to account for text that flows past the edges
        tl = self.style['tickLength']
        if self.orientation == 'left':
            rect = rect.adjusted(0, -m, -min(0,tl), m)
        elif self.orientation == 'right':
            rect = rect.adjusted(min(0,tl), -m, 0, m)
        elif self.orientation == 'top':
            rect = rect.adjusted(-m, 0, m, -min(0,tl))
        elif self.orientation == 'bottom':
            rect = rect.adjusted(-m, min(0,tl), m, 0)
        return rect

    def shape(self):
        # override shape() to exclude grid lines from getting mouse events
        rect = self.mapRectFromParent(self.geometry())
        path = QtGui.QPainterPath()
        path.addRect(rect)
        return path

    def paint(self, p, opt, widget):
        profiler = debug.Profiler()
        if self.picture is None:
            try:
                picture = QtGui.QPicture()
                painter = QtGui.QPainter(picture)
                if self.style["tickFont"]:
                    painter.setFont(self.style["tickFont"])
                specs = self.generateDrawSpecs(painter)
                profiler('generate specs')
                if specs is not None:
                    self.drawPicture(painter, *specs)
                    profiler('draw picture')
            finally:
                painter.end()
            self.picture = picture
        self.picture.play(p)


    def setTickDensity(self, density=1.0):
        """
        Set the density of ticks displayed on the axis.

        A higher density value means that more ticks will be displayed. The density
        value is used in conjunction with the tickSpacing method to determine the
        actual tick locations.

        Parameters
        ----------
        density : float, optional
            Density of ticks to display, by default 1.0.
        """
        self._tickDensity = density
        self.picture = None
        self.update()


    def setTicks(
        self,
        ticks: list[list[tuple[float, str]]] | None
    ):
        """
        Explicitly determine which ticks to display.

        This overrides the behavior specified by
        :meth:`~pyqtgraph.AxisItem.tickSpacing`, :meth:`~pyqtgraph.AxisItem.tickValues`,
        and :meth:`~pyqtgraph.AxisItem.tickStrings`.
        
        The format for *ticks* looks like::

            [
                [
                    (majorTickValue1, majorTickString1),
                    (majorTickValue2, majorTickString2),
                    ...
                ],
                [
                    (minorTickValue1, minorTickString1),
                    (minorTickValue2, minorTickString2),
                    ...
                ],
                ...
            ]
        
        The two levels of major and minor ticks are expected. A third tier of additional
        ticks is optional. If *ticks* is ``None``, then the default tick system will be
        used.

        Parameters
        ----------
        ticks : list of list of float, str or None
            Explicitly set tick display information.
        
        See Also
        --------
        :meth:`~pyqtgraph.AxisItem.tickSpacing`
            How tick spacing is configured.
        :meth:`~pyqtgraph.AxisItem.tickValues`
            How tick values are set.
        :meth:`~pyqtgraph.AxisItem.tickStrings`
            How tick strings are specified.
        """        

        self._tickLevels = ticks
        self.picture = None
        self.update()

    def setTickSpacing(
        self,
        major: float | None=None,
        minor: float | None=None,
        levels: list[tuple[float, float]] | None=None
    ):
        """
        Explicitly determine the spacing of major and minor ticks.

        This overrides the default behavior of the tickSpacing method, and disables
        the effect of setTicks(). Arguments may be either *major* and *minor*,
        or *levels* which is a list of ``(spacing, offset)`` tuples for each
        tick level desired. If no arguments are given, then the default
        behavior of tickSpacing is enabled.

        Parameters
        ----------
        major : float, optional
            Spacing for major ticks, by default None.
        minor : float, optional
            Spacing for minor ticks, by default None.
        levels : list of tuple of float, float, optional
            A list of (spacing, offset) tuples for each tick level, by default None.

        Examples
        --------
        .. code-block:: python

            # two levels, all offsets = 0
            axis.setTickSpacing(5., 1.)
            # three levels, all offsets = 0
            axis.setTickSpacing(levels=[(3., 0.), (1., 0.), (0.25, 0.)])
            # reset to default
            axis.setTickSpacing()
        """

        if levels is None:
            levels = None if major is None else [(major, 0.), (minor, 0.)]
        self._tickSpacing = levels
        self.picture = None
        self.update()

    def tickSpacing(self, minVal: float, maxVal: float, size: float):
        """
        Determine the spacing of ticks on the axis.

        This method is called whenever the axis needs to be redrawn and is a
        good method to override in subclasses that require control over tick locations.

        Parameters
        ----------
        minVal : float
            Minimum value being displayed on the axis.
        maxVal : float
            Maximum value being displayed on the axis.
        size : float
            Length of the axis in pixels.

        Returns
        -------
        list of tuple of float, float
            A list of tuples, one for each tick level.
            Each tuple contains two values: ``(spacing, offset)``.  The spacing value
            is the distance between ticks, and the offset is the first tick relative to
            *minVal*. For example, if ``result[0]`` is ``(10, 0)``, then major ticks 
            will be displayed every 10 units and the first major tick will correspond to
            ``minVal``. If instead ``result[0]`` is ``(10, 5)``, then major ticks will
            be displayed every 10 units, but the first major tick will correspond to
            ``minVal + 5``.

            .. code-block:: python

                [
                    (major_tick_spacing, offset),
                    (minor_tick_spacing, offset),
                    (sub_minor_tick_spacing, offset),
                    ...
                ]
        """
        # First check for explicit tick spacing
        if self._tickSpacing is not None:
            return self._tickSpacing

        dif = abs(maxVal - minVal)
        if dif == 0:
            return []

        ref_size = 300. # axes longer than this display more than the minimum number of major ticks
        minNumberOfIntervals = max(
            2.25,       # 2.0 ensures two tick marks. Fudged increase to 2.25 allows room for tick labels.
            2.25 * self._tickDensity * sqrt(size/ref_size) # sub-linear growth of tick spacing with size
        )

        majorMaxSpacing = dif / minNumberOfIntervals

        # We want to calculate the power of 10 just below the maximum spacing.
        # Then divide by ten so that the scale factors for subdivision all become intergers.
        # p10unit = 10**( floor( log10(majorMaxSpacing) ) ) / 10

        # And we want to do it without a log operation:
        mantissa, exp2 = frexp(majorMaxSpacing) # IEEE 754 float already knows its exponent, no need to calculate
        p10unit = 10. ** ( # approximate a power of ten base factor just smaller than the given number
            floor(            # int would truncate towards zero to give wrong results for negative exponents
                (exp2-1)      # IEEE 754 exponent is ceiling of true exponent --> estimate floor by subtracting 1
                / 3.32192809488736 # division by log2(10)=3.32 converts base 2 exponent to base 10 exponent
            ) - 1             # subtract one extra power of ten so that we can work with integer scale factors >= 5
        )
        # neglecting the mantissa can underestimate by one power of 10 when the true value is JUST above the threshold.
        if 100. * p10unit <= majorMaxSpacing: # Cheaper to check this than to use a more complicated approximation.
            majorScaleFactor = 10
            p10unit *= 10.
        else:
            for majorScaleFactor in (50, 20, 10):
                if majorScaleFactor * p10unit <= majorMaxSpacing:
                    break # find the first value that is smaller or equal
        majorInterval = majorScaleFactor * p10unit
        # manual sanity check: print(f"{majorMaxSpacing:.2e} > {majorInterval:.2e} = {majorScaleFactor:.2e} x {p10unit:.2e}")
        levels = [
            (majorInterval, 0),
        ]

        if self.style['maxTickLevel'] >= 1:
            minorMinSpacing = 2 * dif/size   # no more than one minor tick per two pixels
            trials = (5, 10) if majorScaleFactor == 10 else (10, 20, 50)
            for minorScaleFactor in trials:
                minorInterval = minorScaleFactor * p10unit
                if minorInterval >= minorMinSpacing:
                    break # find the first value that is larger or equal to allowed minimum of 1 per 2px
            levels.append((minorInterval, 0))

        # extra ticks at 10% of major interval are pretty, but eat up CPU
        if self.style['maxTickLevel'] >= 2: # consider only when enabled
            if majorScaleFactor == 10:
                trials = (1, 2, 5, 10) # start at 10% of major interval, increase if needed
            elif majorScaleFactor == 20:
                trials = (2, 5, 10, 20) # start at 10% of major interval, increase if needed
            elif majorScaleFactor == 50:
                trials = (5, 10, 50) # start at 10% of major interval, increase if needed
            else: # invalid value
                trials = () # skip extra interval
                extraInterval = minorInterval
            for extraScaleFactor in trials:
                extraInterval = extraScaleFactor * p10unit
                if extraInterval >= minorMinSpacing or extraInterval == minorInterval:
                    break # find the first value that is larger or equal to allowed minimum of 1 per 2px
            if extraInterval < minorInterval: # add extra interval only if it is visible
                levels.append((extraInterval, 0))
        return levels


    def tickValues(self, minVal:float, maxVal:float, size: float):
        """
        Return the values and spacing of ticks to draw.

        The values returned are essentially the same as those returned by
        :meth:`~pyqtgraph.AxisItem.tickSpacing`, but with the addition of
        explicit tick values for each tick level. This method is a good
        method to override in subclasses.

        Parameters
        ----------
        minVal : float
            Minimum value to generate tick values for.
        maxVal : float
            Maximum value to generate tick values for.
        size : float
            The length of the axis in pixels.

        Returns
        -------
        list of tuple of float, list of float
            A list of tuples, one for each tick level. Each tuple contains two
            values: ``(spacing, values)``, where *spacing* is the distance between
            ticks and *values* is a list of tick values.
        """
        minVal, maxVal = sorted((minVal, maxVal))

        minVal *= self.scale
        maxVal *= self.scale

        ticks = []
        tickLevels = self.tickSpacing(minVal, maxVal, size)
        allValues = np.array([])
        for i in range(len(tickLevels)):
            spacing, offset = tickLevels[i]

            ## determine starting tick
            start = (ceil((minVal-offset) / spacing) * spacing) + offset

            ## determine number of ticks
            num = int((maxVal-start) / spacing) + 1
            values = (np.arange(num) * spacing + start) / self.scale
            ## remove any ticks that were present in higher levels
            ## we assume here that if the difference between a tick value and a previously seen tick value
            ## is less than spacing/100, then they are 'equal' and we can ignore the new tick.
            close = np.any(
                np.isclose(
                    allValues,
                    values[:, np.newaxis],
                    rtol=0,
                    atol=spacing/self.scale*0.01
                ),
                axis=-1
            )
            values = values[~close]
            allValues = np.concatenate([allValues, values])
            ticks.append((spacing/self.scale, values.tolist()))
        if self.logMode:
            return self.logTickValues(minVal, maxVal, size, ticks)
        return ticks

    def logTickValues(self, minVal, maxVal, size, stdTicks):
        """
        Return tick values for log-scale axes.

        This method is called by :meth:`~pyqtgraph.AxisItem.tickValues` when the axis
        is in logarithmic mode. It is a good method to override in subclasses.

        Parameters
        ----------
        minVal : float
            Minimum value to generate tick values for.
        maxVal : float
            Maximum value to generate tick values for.
        size : float
            The length of the axis in pixels.
        stdTicks : list of tuple of float, float
            The tick values generated by the standard
            :meth:`~pyqtgraph.AxisItem.tickValues` method.

        Returns
        -------
        list of tuple of float, float or list of tuple of None, float
            A list of tuples, one for each tick level. Each tuple contains two
            values: ``(spacing, values)``, where *spacing* is the distance between
            ticks and *values* is a list of tick values.
        """

        ## start with the tick spacing given by tickValues().
        ## Any level whose spacing is < 1 needs to be converted to log scale
        ticks = [(spacing, t) for spacing, t in stdTicks if spacing >= 1.0]
        if len(ticks) < 3:
            v1 = int(floor(minVal))
            v2 = int(ceil(maxVal))

            # minor = [v + np.log10(np.arange(1, 10)) for v in range(v1, v2)]
            minor = []
            for v in range(v1, v2):
                minor.extend(v + np.log10(np.arange(1, 10)))
            minor = [x for x in minor if x > minVal and x < maxVal]
            ticks.append((None, minor))
        return ticks

    def tickStrings(self, values: list[float], scale: float, spacing: float):
        """
        Return the strings that should be displayed at each tick value.

        This method is used to generate tick strings, and is called automatically.

        Parameters
        ----------
        values : list of float
            List of tick values.
        scale : float
            The scaling factor for tick values.
        spacing : float
            The spacing between ticks.

        Returns
        -------
        list of str
            List of strings to display at each tick value.
        """

        if self.logMode:
            return self.logTickStrings(values, scale, spacing)

        places = max(0, ceil(-log10(spacing * scale)))
        strings = []
        for v in values:
            vs = v * scale
            if abs(vs) < .001 or abs(vs) >= 10000:
                vstr = "%g" % vs
            else:
                vstr = ("%%0.%df" % places) % vs
            strings.append(vstr)
        return strings

    def logTickStrings(self, values: list[float], scale: float, spacing: float):
        """
        Return the strings that should be displayed at each tick value in log mode.

        This method is called by :meth:`~pyqtgraph.AxisItem.tickStrings` when the axis
        is in logarithmic mode. It is a good method to override in subclasses.

        Parameters
        ----------
        values : list of float
            List of tick values.
        scale : float
            The scaling factor for tick values.
        spacing : float
            The spacing between ticks.

        Returns
        -------
        list of str
            List of strings to display at each tick value.
        """
        estrings = [
            "%0.1g"%x
            for x in 10 ** np.array(values).astype(float) * np.array(scale)
        ]
        convdict = {"0": "⁰",
                    "1": "¹",
                    "2": "²",
                    "3": "³",
                    "4": "⁴",
                    "5": "⁵",
                    "6": "⁶",
                    "7": "⁷",
                    "8": "⁸",
                    "9": "⁹",
                    }
        dstrings = []
        for e in estrings:
            if e.count("e"):
                v, p = e.split("e")
                sign = "⁻" if p[0] == "-" else ""
                pot = "".join([convdict[pp] for pp in p[1:].lstrip("0")])
                v = "" if v == "1" else f"{v}·"
                dstrings.append(f"{v}10{sign}{pot}")
            else:
                dstrings.append(e)
        return dstrings

    def generateDrawSpecs(self, p):
        """
        Generate the drawing specifications for the axis, ticks, and labels.

        This method determines all the coordinates and other information needed to draw
        the axis, including tick positions, tick labels, and axis label. It returns a
        tuple of values that are used to draw the axis. This is a good method to
        override in subclasses that need more control over the appearance of the axis.

        Parameters
        ----------
        p : QPainter
            The painter used to draw the axis.

        Returns
        -------
        tuple
            A tuple containing the drawing specifications for the axis, ticks, and
            labels. The tuple contains the following values:

            - ``axisSpec``: A tuple containing the pen, start point, and end point of
              the axis line.
            - ``tickSpecs``: A list of tuples, one for each tick. Each tuple contains
              the pen, start point, and end point of the tick line.
            - ``textSpecs``: A list of tuples, one for each tick label. Each tuple
              contains the bounding rectangle, alignment flags, and text of the label.
        
        :meta private:
        """
        profiler = debug.Profiler()
        if self.style['tickFont'] is not None:
            p.setFont(self.style['tickFont'])
        bounds = self.mapRectFromParent(self.geometry())

        linkedView = self.linkedView()
        if linkedView is None or self.grid is False:
            tickBounds = bounds
        else:
            tickBounds = linkedView.mapRectToItem(self, linkedView.boundingRect())

        left_offset = -1.0
        right_offset = 1.0
        top_offset = -1.0
        bottom_offset = 1.0
        if self.orientation == 'left':
            span = (bounds.topRight() + Point(left_offset, top_offset),
                    bounds.bottomRight() + Point(left_offset, bottom_offset))
            tickStart = tickBounds.right()
            tickStop = bounds.right()
            tickDir = -1
            axis = 0
        elif self.orientation == 'right':
            span = (bounds.topLeft() + Point(right_offset, top_offset),
                    bounds.bottomLeft() + Point(right_offset, bottom_offset))
            tickStart = tickBounds.left()
            tickStop = bounds.left()
            tickDir = 1
            axis = 0
        elif self.orientation == 'top':
            span = (bounds.bottomLeft() + Point(left_offset, top_offset),
                    bounds.bottomRight() + Point(right_offset, top_offset))
            tickStart = tickBounds.bottom()
            tickStop = bounds.bottom()
            tickDir = -1
            axis = 1
        elif self.orientation == 'bottom':
            span = (bounds.topLeft() + Point(left_offset, bottom_offset),
                    bounds.topRight() + Point(right_offset, bottom_offset))
            tickStart = tickBounds.top()
            tickStop = bounds.top()
            tickDir = 1
            axis = 1
        else:
            raise ValueError(
                "self.orientation must be in {'left', 'right', 'top', 'bottom'}"
            )
        ## determine size of this item in pixels
        points = list(map(self.mapToDevice, span))
        if None in points:
            return
        lengthInPixels = Point(points[1] - points[0]).length()
        if lengthInPixels == 0:
            return

        # Determine major / minor / subminor axis ticks
        if self._tickLevels is None:
            tickLevels = self.tickValues(self.range[0], self.range[1], lengthInPixels)
            tickStrings = None
        else:
            ## parse self.tickLevels into the formats returned by tickLevels() and tickStrings()
            tickLevels = []
            tickStrings = []
            for level in self._tickLevels:
                values = []
                strings = []
                tickLevels.append((None, values))
                tickStrings.append(strings)
                for val, strn in level:
                    values.append(val)
                    strings.append(strn)

        ## determine mapping between tick values and local coordinates
        dif = self.range[1] - self.range[0]
        if dif == 0:
            xScale = 1
            offset = 0
        elif axis == 0:
            xScale = -bounds.height() / dif
            offset = self.range[0] * xScale - bounds.height()
        else:
            xScale = bounds.width() / dif
            offset = self.range[0] * xScale

        xRange = [x * xScale - offset for x in self.range]
        xMin = min(xRange)
        xMax = max(xRange)

        profiler('init')

        tickPositions = [] # remembers positions of previously drawn ticks

        ## compute coordinates to draw ticks
        ## draw three different intervals, long ticks first
        tickSpecs = []
        for i in range(len(tickLevels)):
            tickPositions.append([])
            ticks = tickLevels[i][1]

            ## length of tick
            tickLength = self.style['tickLength'] / ((i*0.5)+1.0)

            lineAlpha = self.style["tickAlpha"]
            if lineAlpha is None:
                lineAlpha = 255 / (i+1)
                if self.grid is not False:
                    lineAlpha *= self.grid/255. * fn.clip_scalar((0.05  * lengthInPixels / (len(ticks)+1)), 0., 1.)
            elif isinstance(lineAlpha, float):
                lineAlpha *= 255
                lineAlpha = max(0, int(round(lineAlpha)))
                lineAlpha = min(255, int(round(lineAlpha)))
            elif isinstance(lineAlpha, int):
                if (lineAlpha > 255) or (lineAlpha < 0):
                    raise ValueError("lineAlpha should be [0..255]")
            else:
                raise TypeError("Line Alpha should be of type None, float or int")
            tickPen = self.tickPen()
            if tickPen.brush().style() == QtCore.Qt.BrushStyle.SolidPattern: # only adjust simple color pens
                tickPen = QtGui.QPen(tickPen) # copy to a new QPen
                color = QtGui.QColor(tickPen.color()) # copy to a new QColor
                color.setAlpha(int(lineAlpha)) # adjust opacity
                tickPen.setColor(color)

            for v in ticks:
                ## determine actual position to draw this tick
                x = (v * xScale) - offset
                if x < xMin or x > xMax:  ## last check to make sure no out-of-bounds ticks are drawn
                    tickPositions[i].append(None)
                    continue
                tickPositions[i].append(x)

                p1 = [x, x]
                p2 = [x, x]
                p1[axis] = tickStart
                p2[axis] = tickStop
                if self.grid is False:
                    p2[axis] += tickLength*tickDir
                tickSpecs.append((tickPen, Point(p1), Point(p2)))
        profiler('compute ticks')


        if self.style['stopAxisAtTick'][0] is True:
            minTickPosition = min(map(min, tickPositions))
            if axis == 0:
                stop = max(span[0].y(), minTickPosition)
                span[0].setY(stop)
            else:
                stop = max(span[0].x(), minTickPosition)
                span[0].setX(stop)
        if self.style['stopAxisAtTick'][1] is True:
            maxTickPosition = max(map(max, tickPositions))
            if axis == 0:
                stop = min(span[1].y(), maxTickPosition)
                span[1].setY(stop)
            else:
                stop = min(span[1].x(), maxTickPosition)
                span[1].setX(stop)
        axisSpec = (self.pen(), span[0], span[1])


        textOffset = self.style['tickTextOffset'][axis]  ## spacing between axis and text
        textSize2 = 0
        lastTextSize2 = 0
        textRects = []
        textSpecs = []  ## list of draw

        # If values are hidden, return early
        if not self.style['showValues']:
            return (axisSpec, tickSpecs, textSpecs)

        for i in range(min(len(tickLevels), self.style['maxTextLevel']+1)):
            ## Get the list of strings to display for this level
            if tickStrings is None:
                spacing, values = tickLevels[i]
                strings = self.tickStrings(values, self.autoSIPrefixScale * self.scale, spacing)
            else:
                strings = tickStrings[i]

            if len(strings) == 0:
                continue

            ## ignore strings belonging to ticks that were previously ignored
            for j in range(len(strings)):
                if tickPositions[i][j] is None:
                    strings[j] = None

            ## Measure density of text; decide whether to draw this level
            rects = []
            for s in strings:
                if s is None:
                    rects.append(None)
                else:
                    br = p.boundingRect(QtCore.QRectF(0, 0, 100, 100), QtCore.Qt.AlignmentFlag.AlignCenter, s)
                    ## boundingRect is usually just a bit too large
                    ## (but this probably depends on per-font metrics?)
                    br.setHeight(br.height() * 0.8)

                    rects.append(br)
                    textRects.append(rects[-1])

            if textRects:
                ## measure all text, make sure there's enough room
                if axis == 0:
                    textSize = np.sum([r.height() for r in textRects])
                    textSize2 = np.max([r.width() for r in textRects])
                else:
                    textSize = np.sum([r.width() for r in textRects])
                    textSize2 = np.max([r.height() for r in textRects])
            else:
                textSize = 0
                textSize2 = 0

            if i > 0:  ## always draw top level
                ## If the strings are too crowded, stop drawing text now.
                ## We use three different crowding limits based on the number
                ## of texts drawn so far.
                textFillRatio = float(textSize) / lengthInPixels
                finished = False
                for nTexts, limit in self.style['textFillLimits']:
                    if len(textSpecs) >= nTexts and textFillRatio >= limit:
                        finished = True
                        break
                if finished:
                    break

            lastTextSize2 = textSize2

            # Determine exactly where tick text should be drawn
            for j in range(len(strings)):
                vstr = strings[j]
                if vstr is None: ## this tick was ignored because it is out of bounds
                    continue
                x = tickPositions[i][j]
                textRect = rects[j]
                height = textRect.height()
                width = textRect.width()
                offset = max(0,self.style['tickLength']) + textOffset

                rect = QtCore.QRectF()
                if self.orientation == 'left':
                    alignFlags = QtCore.Qt.AlignmentFlag.AlignRight|QtCore.Qt.AlignmentFlag.AlignVCenter
                    rect = QtCore.QRectF(tickStop-offset-width, x-(height/2), width, height)
                elif self.orientation == 'right':
                    alignFlags = QtCore.Qt.AlignmentFlag.AlignLeft|QtCore.Qt.AlignmentFlag.AlignVCenter
                    rect = QtCore.QRectF(tickStop+offset, x-(height/2), width, height)
                elif self.orientation == 'top':
                    alignFlags = QtCore.Qt.AlignmentFlag.AlignHCenter|QtCore.Qt.AlignmentFlag.AlignBottom
                    rect = QtCore.QRectF(x-width/2., tickStop-offset-height, width, height)
                elif self.orientation == 'bottom':
                    alignFlags = QtCore.Qt.AlignmentFlag.AlignHCenter|QtCore.Qt.AlignmentFlag.AlignTop
                    rect = QtCore.QRectF(x-width/2., tickStop+offset, width, height)

                textFlags = alignFlags | QtCore.Qt.TextFlag.TextDontClip
                br = self.boundingRect()

                # br.contains(rect) suffers from floating point rounding errors
                if br & rect != rect:
                    continue

                textSpecs.append((rect, textFlags, vstr))
        profiler('compute text')

        ## update max text size if needed.
        self._updateMaxTextSize(lastTextSize2)

        return axisSpec, tickSpecs, textSpecs

    def drawPicture(self, p, axisSpec, tickSpecs, textSpecs):
        profiler = debug.Profiler()

        p.setRenderHint(p.RenderHint.Antialiasing, False)
        p.setRenderHint(p.RenderHint.TextAntialiasing, True)

        ## draw long line along axis
        pen, p1, p2 = axisSpec
        p.setPen(pen)
        p.drawLine(p1, p2)
        # p.translate(0.5,0)  ## resolves some damn pixel ambiguity

        ## draw ticks
        for pen, p1, p2 in tickSpecs:
            p.setPen(pen)
            p.drawLine(p1, p2)
        profiler('draw ticks')

        # Draw all text
        if self.style['tickFont'] is not None:
            p.setFont(self.style['tickFont'])
        p.setPen(self.textPen())
        bounding = self.boundingRect().toAlignedRect()
        p.setClipRect(bounding)
        for rect, flags, text in textSpecs:
            p.drawText(rect, int(flags), text)

        profiler('draw text')

    def show(self):
        super().show()
        if self.orientation in ['left', 'right']:
            self._updateWidth()
        else:
            self._updateHeight()

    def hide(self):
        super().hide()
        if self.orientation in ['left', 'right']:
            self._updateWidth()
        else:
            self._updateHeight()

    def wheelEvent(self, event):
        lv = self.linkedView()
        if lv is None:
            return
        # Did the event occur inside the linked ViewBox (and not over the axis iteself)?
        if lv.sceneBoundingRect().contains(event.scenePos()):
            event.ignore()
            return
        else:
            # pass event to linked viewbox with appropriate single axis zoom parameter
            if self.orientation in ['left', 'right']:
                lv.wheelEvent(event, axis=1)
            else:
                lv.wheelEvent(event, axis=0)
        event.accept()

    def mouseDragEvent(self, event):
        lv = self.linkedView()
        if lv is None:
            return
        # Did the mouse down event occur inside the linked ViewBox (and not the axis)?
        if lv.sceneBoundingRect().contains(event.buttonDownScenePos()):
            event.ignore()
            return
        # otherwise pass event to linked viewbox with appropriate single axis parameter
        if self.orientation in ['left', 'right']:
            return lv.mouseDragEvent(event, axis=1)
        else:
            return lv.mouseDragEvent(event, axis=0)

    def mouseClickEvent(self, event):
        lv = self.linkedView()
        if lv is None:
            return
        return lv.mouseClickEvent(event)
